# shared 模型处理器
from brave.api.service.components.base import ComponentHandler


class SharedComponentHandler(ComponentHandler):
    def create_files(self, component_data):
        print(f"[Shared] 创建公共组件文件: {component_data}")

    def get_files(self, component_data):
        print(f"[Shared] 获取公共组件文件: {component_data}")

    def update_files(self, component_data):
        print(f"[Shared] 更新公共组件文件: {component_data}")

    def delete_files(self, component_data):
        print(f"[Shared] 删除公共组件文件: {component_data}")


from typing import Callable, Awaitable, Dict, Union, Coroutine
import asyncio
from collections import defaultdict

from brave.api.core.base_event_router import BaseEventRouter
from brave.api.core.event import WatchFileEvent

Callback = Union[Callable[[dict], None], Callable[[dict], Coroutine]]

class WatchFileEvenetRouter(BaseEventRouter[WatchFileEvent,Callback]):
    async def dispatch(self, event: WatchFileEvent, payload: dict):
        handlers = self._handlers.get(event, [])
        for handler in handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler(payload)
            else:
                handler(payload)
    # def __init__(self):
    #     self._handlers: dict[WatchFileEvent, set[Callback]] = defaultdict(set)

    # def on_event(self, event: WatchFileEvent):
    #     def decorator(func: Callback):
    #         self._handlers[event].add(func)
    #         return func
    #     return decorator

    # def register_handler(self, event: WatchFileEvent, handler: Callback):
    #     self._handlers[event].add(handler)


    # async def dispatch(self, event:WatchFileEvent,msg: dict):
    #     # event = msg.get("file_type")
    #     # if not event:
    #     #     print("[WatchFileEvenetRouter] No 'file_type' in message", msg)
    #     #     return
    #     handlers = self._handlers.get(event)
    #     if handlers:
    #         for handler in handlers:
    #             if asyncio.iscoroutinefunction(handler):
    #                 await handler(msg)
    #             else:
    #                 handler(msg)
    #     else:
    #         print(f"[WatchFileEvenetRouter] No handler for event '{event}'")




import json
import pandas as pd 
# def get_data(item):
#     content = item.content #json.loads(item.content)
#     return {
#         "sample_key":item.sample_key,
#         "fastq1":content['fastq1'],
#         "fastq2":content['fastq2'],
#     }

def get_db_field():
    return []


def parse_data(request_param,db_dict):
    # reads = db_dict['clean_reads']
    # samples = [get_data(item) for item in reads]
    # result = {
    #     "samples":samples,
    #     "genome_index":request_param['genome_index']
    # }
    return []



from datetime import datetime
from sqlalchemy import Column, DateTime, Table
from sqlalchemy.sql.sqltypes import Integer, String,Boolean
from brave.api.config.db import meta
# from sqlalchemy.dialects.mysql import LONGTEXT
from sqlalchemy import Text,Index
from sqlalchemy.dialects.mysql import LONGTEXT


t_project = Table(
    "t_project",
    meta,
    Column("id", Integer, primary_key=True),
    Column("project_id", String(255)),
    Column("project_name", String(255)),
    Column("metadata_form", Text)
)

samples = Table(
    "t_samples",
    meta,
    Column("id", Integer, primary_key=True),
    Column("sample_id", String(255)),
    Column("project", String(255)),
    # Column("sample_key", String(255)),
    # Column("analysis_key", String(255), unique=True),
    Column("sample_name", String(255)),
    # Column("sequencing_target", String(255)),
    # Column("sequencing_technique", String(255)),
    # Column("sample_composition", String(255)),
    # Column("library_name", String(255)),
    Column("sample_group", String(255)),
    # Column("sample_group_name", String(255)),
    # Column("sample_source", String(255)),
    # Column("host_disease", String(255)),
    # Column("sample_individual", String(255)),
    # Column("is_available", Integer),
    # Column("fastq1", String(255)),
    # Column("fastq2", String(255)),
    Column("metadata", Text)
)
analysis = Table(
    "nextflow",
    meta,
    Column("id", Integer, primary_key=True),
    Column("project", String(255)),
    Column("analysis_id", String(255)),
    Column("component_id", String(255)),
    Column("analysis_name", String(255)),
    Column("input_file", String(255)),
    Column("analysis_method", String(255)),
    Column("work_dir", String(255)),
    Column("params_path", String(255)),
    Column("command_path", String(255)),
    Column("request_param", Text().with_variant(LONGTEXT(), "mysql")),
    Column("output_format", Text().with_variant(LONGTEXT(), "mysql")),
    Column("output_dir", String(255)),
    Column("pipeline_script", String(255)),
    Column("parse_analysis_module", String(255)),
    Column("trace_file", String(255)),
    Column("workflow_log_file", String(255)),
    Column("executor_log_file", String(255)),
    Column("process_id", String(255)),
    Column("script_config_file", String(255)),
    # Column("analysis_status", String(255)),
    Column("job_id", String(255)),
    Column("ports", String(255)),
    Column("url", String(255)),
    # Column("run_type", String(255)),
    # Column("job_id", String(255)),
    # Column("server_id", String(255)),
    Column("job_status", String(255)),
    Column("server_status", String(255)),


    Column("command_log_path", String(255)),
    Column("is_report", Boolean, default=False),

    # Column("container_id", String(255)),
    # Column("sub_container_id", String(255)),
    Column("data_component_ids",Text),
    Column("extra_project_ids",Text().with_variant(LONGTEXT(), "mysql"))
)

analysis_result = Table(
    "analysis_result",
    meta,
    Column("id", Integer, primary_key=True),
    Column("analysis_result_id", String(255)),
    Column("sample_id", String(255)),
    Column("sample_source", String(255)),


    # Column("sample_name", String(255)),
    # Column("sample_key", String(255)),
    # Column("analysis_name", String(255)),
    Column("file_name", String(255)),
    Column("analysis_key", String(255)),
    Column("component_id", String(255)),
    # Column("analysis_method", String(255)),
    Column("software", String(255)),
    Column("content",Text().with_variant(LONGTEXT(), "mysql")),
    Column("analysis_version", String(255)),
    Column("content_type", String(255)),
    Column("analysis_id", String(255)),
    Column("project", String(255)),
    Column("request_param", String(255)),
    Column("analysis_type", String(255)),
    Column("create_date", String(255)),
    Column("analysis_result_hash", String(255))
)

literature = Table(
    "literature",
    meta,
    Column("id", Integer, primary_key=True),
    Column("literature_key", String(255)),
    Column("literature_type", String(255)),
    Column("title", String(255)),
    Column("url", String(255)),
    Column("content", Text),
    Column("translate", Text),
    Column("interpretation", Text),
    Column("img", Text),
    Column("journal", String(255)),
    Column("publish_date", String(255)),
    Column("keywords", String(255))

)


relation_literature = Table(
    "relation_literature",
    meta,
    Column("relation_id", Integer, primary_key=True),
    Column("literature_key", String(255)),
    Column("obj_key", String(255)),
    Column("obj_type", String(255))
)

# pipeline_type: pipelne analysis_software analysis_file  script_analysis
# 

# Workflow
#    |
#    |  (多个)
#    v
# WorkflowPipelineRelation  <-- ordered DAG / parallel execution
#    |
#    |  (1 对 1)
#    v
# Pipeline
#    |
#    +--> InputFiles / OutputFiles / Downstream
# pipeline,software,file,downstream
t_pipeline_components = Table(
    "pipeline_components",
    meta,
    Column("id", Integer, primary_key=True),
    Column("component_id", String(255)),
    Column("install_key", String(255)),
    Column("component_type", String(255)), 
    Column("component_name", String(255)), 
    Column("description", Text().with_variant(LONGTEXT(), "mysql")), 
    Column("img", String(255)), 
    Column("container_id", String(255)),
    Column("sub_container_id", String(255)),
    Column("tags", String(255)), 
    Column("category", String(255)), 
    # Column("namespace", String(255)),
    Column("content", Text),
    Column("order_index", Integer),
    Column("position", Text),
    Column("edges", Text)

)
# relation_type: pipeline_software software_input_file  software_ouput_file  file_script
t_pipeline_components_relation = Table(
    "pipeline_components_relation",
    meta,
    Column("id", Integer, primary_key=True),
    Column("relation_id", String(255)),
    Column("relation_type", String(255)), 
    Column("install_key", String(255)),
    # Column("pipeline_id", String(255)),
    Column("component_id", String(255)),
    Column("parent_component_id", String(255)),
    Column("order_index", Integer),
    # Column("namespace", String(255)),
    Column("created_at", DateTime, default=datetime.now),
    Column("updated_at", DateTime, onupdate=datetime.now)

)

t_pipeline_components_edges = Table(
    "pipeline_components_edges",
    meta,
    Column("id", Integer, primary_key=True),
    Column("edge_id", String(255)),
    Column("source", String(255)),
    Column("sourceHandle", String(255)),
    Column("target", String(255)),
    Column("targetHandle", String(255)),
    Column("pipeline_id", String(255))
)

t_bio_database = Table(
    "bio_database",
    meta,
    Column("id", Integer, primary_key=True),
    Column("database_id", String(255)),
    Column("name", String(255)),
    Column("path", String(255)),
    Column("type", String(255)),
    Column("db_index", String(255))
)

t_namespace = Table(
    "t_namespace",
    meta,
    Column("id", Integer, primary_key=True),
    Column("namespace_id", String(255)),
    Column("name", String(255)),
    Column("volumes", Text().with_variant(LONGTEXT(), "mysql")),
    Column("is_use",  Boolean, default=False),
)

# t_relation_pipeline_software = Table(
#     "relation_pipeline_software",
#     meta,
#     Column("relation_id", Integer, primary_key=True),
#     Column("pipeline_id", String(255)),
#     Column("analysis_software_id", String(255))
# )

# t_analysis_software = Table(
#     "analysis_software",
#     meta,
#     Column("id", Integer, primary_key=True),
#     Column("analysis_software_id", String(255)),
#     Column("content", Text)

# )
# t_relation_software_file = Table(
#     "relation_software_file",
#     meta,
#     Column("relation_id", Integer, primary_key=True),
#     Column("analysis_software_id", String(255)),
#     Column("analysis_file_id", String(255)),
#     Column("file_type", String(255)),
#     Column("content", Text)
# )


# t_analysis_file = Table(
#     "analysis_file",
#     meta,
#     Column("id", Integer, primary_key=True),
#     Column("analysis_file_id", String(255))

# )
# t_relation_file_script = Table(
#     "relation_file_script",
#     mata,
#     Column("relation_id", Integer, primary_key=True),
#     Column("analysis_file_id", String(255)),
#     Column("downstream_analysis_id", String(255))
# )
# t_downstream_analysis = Table(
#     "downstream_analysis",
#     meta,
#     Column("id", Integer, primary_key=True),
#     Column("downstream_analysis_id", String(255))
# )
# # meta.create_all(engine)



# t_application = Table(
#     "application",
#     meta,
#     Column("id", Integer, primary_key=True),
#     Column("application_id", String(255)),
#     Column("name", String(255)),
#     Column("image", String(255)),
#     Column("description", String(255)),
#     Column("created_at", DateTime, default=datetime.now),
#     Column("updated_at", DateTime, onupdate=datetime.now)
# )

t_container = Table(
    "container",
    meta,
    Column("id", Integer, primary_key=True),
    Column("container_id", String(255)),
    Column("container_key", String(255)),
    Column("name", String(255)),
    Column("image", String(255)),
    Column("img",  Text().with_variant(LONGTEXT(), "mysql")),
    Column("image_id", String(255)),
    Column("image_status", String(255)),
    Column("description", String(255)),
    # Column("namespace", String(255)),
    Column("envionment", String(255)),
    Column("command", String(255)),
    Column("port", String(255)),
    Column("labels", Text().with_variant(LONGTEXT(), "mysql")),
    Column("change_uid", Boolean, default=True),
    Column("created_at", DateTime, default=datetime.now),
    Column("updated_at", DateTime, onupdate=datetime.now)
)




from operator import and_, or_
from shortuuid import uuid
from brave.microbe.schemas.entity import PageEntity
from brave.microbe.models.core import t_disease
from sqlalchemy import select,func
from sqlalchemy.engine import Connection
import pandas as pd
from sqlalchemy import select,func,insert,case,exists

from brave.microbe.utils import import_ncbi_mesh
import json
def page_disease(conn: Connection, query: PageEntity):
    if query.parent_id and query.parent_id=="default":
        query.parent_id = "F03"
    # stmt =select(
    #     t_disease,
    # ) 
    child = t_disease.alias("child")
    stmt = (
        select(
            t_disease,
            case(
                (exists().where(child.c.parent_entity_id == t_disease.c.entity_id), True),
                else_=False
            ).label("has_children")
        )
    )
    
    conditions = []
  
    if query.keywords:
        keyword_pattern = f"%{query.keywords.strip()}%"
        conditions.append(
             t_disease.c.entity_name.ilike(keyword_pattern)
        )
        
    if query.parent_id:
        conditions.append(
            t_disease.c.parent_entity_id == query.parent_id
        )
    
    
    if conditions:  # 只有有条件时才加 where
        conditions = and_(*conditions) if len(conditions) > 1 else conditions[0]
        stmt = stmt.where(conditions)
        count_stmt = select(func.count()).select_from(t_disease).where(conditions)
    else:
        count_stmt = select(func.count()).select_from(t_disease)
    if query.page_size != -1:
        stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_disease = conn.execute(stmt).mappings().all()
    total = conn.execute(count_stmt).scalar()

    return {
        "items": find_disease,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }

def find_disease_by_id(conn: Connection, entity_id: str):
    stmt = t_disease.select().where(t_disease.c.entity_id ==entity_id)
    find_disease = conn.execute(stmt).mappings().first()
    return find_disease

def update_disease(conn: Connection, entity_id: str, updateData: dict):
    stmt = t_disease.update().where(t_disease.c.entity_id == entity_id).values(updateData)
    conn.execute(stmt)

def add_disease(conn: Connection, data: dict):
    data['entity_id'] = str(uuid())
    stmt = t_disease.insert().values(data)
    conn.execute(stmt)

def find_disease_by_keywords(conn: Connection, keywords: str):
    keyword_pattern = f"%{keywords.strip()}%"
    stmt = t_disease.select().where(t_disease.c.entity_name.ilike(keyword_pattern)).limit(10)
    find_disease = conn.execute(stmt).mappings().all()
    return find_disease

def details_disease_by_id(conn: Connection, entity_id: str):
    stmt = t_disease.select().where(t_disease.c.entity_id == entity_id)
    details_disease = conn.execute(stmt).mappings().first()
    return details_disease

def import_diseases(conn: Connection, records: list, batch_size: int = 1000):
    inserted = 0
   
    for i in range(0, len(records), batch_size):
        batch = records[i:i + batch_size]
        stmt = t_disease.insert()
        conn.execute(stmt, batch)
        inserted += len(batch)
        print(f"Inserted {inserted} records...")

    return {"message": f"导入完成，共导入 {inserted} 行"}

def delete_disease_by_id(conn: Connection, entity_id: str):
    stmt = t_disease.delete().where(t_disease.c.entity_id == entity_id)
    conn.execute(stmt)


def get_parent(tree_number):
    if pd.isna(tree_number):
        return None
    parts = tree_number.split(".")
    return ".".join(parts[:-1]) if len(parts) > 1 else None

def init(conn: Connection):
    json_data = import_ncbi_mesh.get_json("/ssd1/wy/workspace2/nextflow-fastapi/desc2025.xml")
    df = pd.DataFrame(json_data)
    df_exploded = df.explode(["TreeNumberList"], ignore_index=True)

    df_exploded["ParentTree"] = df_exploded["TreeNumberList"].apply(get_parent)
    df_exploded = df_exploded[pd.notna(df_exploded["TreeNumberList"])]

    df_f03 = df_exploded[df_exploded["TreeNumberList"].str.startswith("F03")]
    df_f03 = df_f03.drop_duplicates(subset=["DescriptorUI"], keep="first")
    df_f03_rename = df_f03[["DescriptorUI","DescriptorName","TreeNumberList","ParentTree"]].rename({
        "DescriptorUI":"mesh_id",
        "DescriptorName":"entity_name",
        "TreeNumberList":"entity_id",
        "ParentTree":"parent_entity_id",
    },axis=1)
    data = json.loads(df_f03_rename.to_json(orient="records"))
    return import_diseases(conn, data)


import os
# from brave.api.routes.sse import global_queue
import asyncio
# from fastapi import Depends, FastAPI

import json
from turtle import update

from brave.api.service.analysis_result_parse import AnalysisResultParse
from brave.api.service.sse_service import SSESessionService
# def get_sse_service(app: FastAPI = Depends()):
#     return app.state.sse_service
from brave.api.config.db import get_engine
from brave.api.models.core import analysis as t_analysis
from sqlalchemy import select, update


async def file_change(change,file_path,sse_service:SSESessionService,analysis_result_parse_service:AnalysisResultParse):
    # sse_service = app.state.sse_service
    if "trace" in file_path:
        analysis_id = os.path.basename(file_path).replace(".trace.log","")
        # await asyncio.sleep(2)
        data = json.dumps({
            "analysis_id":analysis_id,
            "msgType":"trace",
            "msg":f"监控分析{analysis_id}文件{file_path}发生变化!"
        })
        print(f"监控分析{analysis_id}文件{file_path}发生变化!")

        msg = {"group": "default", "data": data}
        await sse_service.push_message(msg)
        await analysis_result_parse_service.add_change_analysis_id(analysis_id)
        # await analysis_result_parse_service.auto_save_analysis_result(analysis_id)
    elif "workflow" in file_path:
        analysis_id = os.path.basename(file_path).replace(".workflow.log","")
        # await asyncio.sleep(2)
        data = json.dumps({
            "analysis_id":analysis_id,
            "msgType":"workflow_log"
        })
        msg = {"group": "default", "data": data}
        await sse_service.push_message(msg)
    pass

async def process_end(analysis,sse_service,analysis_result_parse_service:AnalysisResultParse):
    analysis_id = analysis.get("analysis_id")
    with get_engine().begin() as conn:  
        stmt = (
            update(t_analysis)
            .where(t_analysis.c.analysis_id == analysis_id)
            .values(process_id=None,analysis_status = "finished")
        )
        conn.execute(stmt)
        conn.commit()

    # sse_service = app.state.sse_service
    data = json.dumps({
        "analysis_id":analysis.get("analysis_id"),
        "msgType":"process_end",
        "analysis":analysis
    })
    msg = {"group": "default", "data": data}
    await analysis_result_parse_service.add_remove_analysis_id(analysis.get("analysis_id"))
    await sse_service.push_message(msg)



async def analysis_result_add(analysis,analysis_result,sse_service):
    data = json.dumps({
        "msg":f"分析{analysis['analysis_id']}，文件{analysis_result['file_name']}保存成功!",
        "msgType":"analysis_result",
        "component_id":analysis_result['component_id']
    })
    msg = {"group": "default", "data": data}
    await sse_service.push_message(msg)

async def analysis_result_update(analysis,analysis_result,sse_service):
    data = json.dumps({
        "msg":f"分析{analysis['analysis_id']}，文件{analysis_result['file_name']}更新成功!",
        "msgType":"analysis_result",
        "component_id":analysis_result['component_id']
    })
    msg = {"group": "default", "data": data}
    await sse_service.push_message(msg)


import asyncio
from concurrent.futures import ThreadPoolExecutor
import os
from typing import Dict
import docker
from docker.models.containers import Container
from brave.api.core.evenet_bus import EventBus
from brave.api.core.event import AnalysisExecutorEvent
from brave.api.core.routers_name import RoutersName
from brave.api.executor.models import DockerJobSpec
from brave.api.schemas.analysis import AnalysisId
from brave.api.service.analysis_service import find_running_analysis
from .base import JobExecutor
from brave.api.core.routers.workflow_event_router import WorkflowEventRouter    
from brave.api.config.config import get_settings
from brave.api.config.db import get_engine
from docker.errors import NotFound, APIError
import traceback
import brave.api.service.container_service as container_service
from brave.api.config.db import get_engine
import json
from brave.api.schemas.analysis import AnalysisExecuterModal
from docker.errors import ImageNotFound
from brave.api.service import container_service
import os

from brave.api.service import namespace_service


class DockerExecutor(JobExecutor):

    def __init__(self, event_bus: EventBus):
        self.event_bus = event_bus
        self.client = docker.from_env()
        self.containers: Dict[str, Container] = {}
        # self._monitor_task = None
        self._monitor_interval = 2.0  # 秒
        self.to_remove = []
        self.setting  = get_settings()
        # asyncio.create_task(self.recover_running_containers())
        asyncio.create_task(self._monitor_containers())
        asyncio.create_task(self.update_images_status())
        self.running_conntainer =  []
        asyncio.create_task(asyncio.to_thread(self._list_running,recover=True))
        self.executor = ThreadPoolExecutor(max_workers=5)
        


    def recover_running_containers(self,containers):
        """
        程序启动时调用：
        从数据库查询所有运行中分析，恢复对应容器监控
        """
        containers = {container.name:container for container in containers}
        self.containers.update(containers)
        with get_engine().begin() as conn:  
            running_jobs = find_running_analysis(conn)  # 异步获取所有运行中任务
        
        for job in running_jobs:
            try:
                container = self.client.containers.get(job["run_id"])
                self.containers[job["run_id"]] = container
            except Exception as e:
                print(f"Error recovering container {job['run_id']}: {e}")

                self.to_remove.append(job["run_id"])
                pass

        # if self.containers and self._monitor_task is None:
        # self._monitor_task = asyncio.create_task(self._monitor_containers())
    async def _do_submit_job(self, job: AnalysisExecuterModal) :
        # loop = asyncio.get_running_loop()
        # await loop.run_in_executor(
        #     self.executor,
        #     self._sync_submit_job,
        #     job
        # )
        asyncio.create_task(asyncio.to_thread(self._sync_submit_job, job))

        # await asyncio.to_thread(self._sync_submit_job,job)
        pass
        # return container_id
    def is_already_running(self, job_id: str) -> bool:
        try:
            self.client.containers.get(job_id)
            return True
        except NotFound:
            return False




    def _sync_submit_job(self, job: AnalysisExecuterModal) -> str:
        user_id = os.getuid() 
        group_id = os.getgid()
        used_namespace=None
        with get_engine().begin() as conn: 
            find_container = container_service.find_container_by_id(conn,job.container_id)
            if not find_container:
                raise RuntimeError(f"Container {job.container_id} not found")
            used_namespace = namespace_service.get_used_namespace(conn)
    
        # "DISABLE_AUTH":true,
        # "USERID":"$USERID",
        # "GROUPID":"$GROUPID",
        # "R_USER_WORKDIR":"$OUTPUT_DIR",
        # "R_SCRIPT":"$SCRIPT_FILE"
        # }
        sock_gid = os.stat('/var/run/docker.sock').st_gid
        settings = get_settings()
        work_dir = str(settings.WORK_DIR)
        pipeline_dir = str(settings.PIPELINE_DIR)
        base_dir = str(settings.BASE_DIR)
        script_dir = os.path.dirname(job.pipeline_script)
        connom_script_dir = os.path.dirname(script_dir)
        envionment = {}
        if find_container["envionment"]:
            envionment = find_container["envionment"]
            envionment = envionment.replace("$USERID", str(user_id))
            envionment = envionment.replace("$GROUPID", str(group_id))
            envionment = envionment.replace("$DOCKER_GROUPID", str(sock_gid))

            envionment = envionment.replace("$SCRIPT_DIR", script_dir)
            envionment = envionment.replace("$OUTPUT_DIR", job.output_dir)
            envionment = envionment.replace("$SCRIPT_FILE", job.pipeline_script)
            envionment = json.loads(envionment) 
    
        # command = job.command
        # command.extend  (["2>&1","|","tee",f"{job.output_dir}/run.log"])
        # try:
        #     self.client.containers.get(job.job_id)
        #     raise RuntimeError(f"Container {job.job_id} already exists")
        # except NotFound:
        #     pass  # 容器不存在，正常流程
        # except Exception as e:
        #     print(f"Error checking container existence: {e}")
        #     self.to_remove.append(job.job_id)
        #     raise e  # 其他错误不应吞掉



        command = f"-c  \"bash ./run.sh  2>&1 | tee {job.command_log_path}; exit ${{PIPESTATUS[0]}}\""
        port = {}
        docker_uid = user_id
        labels ={}
        network = None
        url_predix = f"container/{job.analysis_id}"
        if job.run_id.startswith("server-") or job.run_id.startswith("retry-"):
            command = find_container["command"] or  ""
            command = command.replace("$SCRIPT_DIR",script_dir)
            command = command.replace("$URL_PREFIX",url_predix)
            # command = f"{command}  > {job.command_log_path}"
            port = {}
            port_str = find_container["port"]
            if ":" in port_str:
                for item in port_str.split(","):
                    host_port, container_port = item.split(":")
                    # Docker SDK 需要字典 {container_port/tcp: host_port}
                    port[f"{container_port}/tcp"] = int(host_port)
            elif  port!="":
                port =  {f"{port_str}/tcp":None}

            docker_uid = user_id if find_container["change_uid"] else None
            labels = find_container["labels"]
            if labels:
                labels = labels.replace("$URL_PREFIX",url_predix)
                labels = labels.replace("$CONTAINER_NAME",job.analysis_id)
                labels = json.loads(labels)
            else:
                labels={}
            
            try:
                network = self.client.networks.get("traefik_proxy")
            except docker.errors.NotFound:
                network = self.client.networks.create("traefik_proxy", driver="bridge")
            network = "traefik_proxy"
            # $SCRIPT_DIR
        volumes = {}

        if used_namespace and used_namespace.volumes:
            volumes_dict = json.loads(used_namespace.volumes)
            for k,v in volumes_dict.items():
                if os.path.exists(k):
                    volumes.update({ k: v})

        # if os.path.exists("/data"):
        #     volumes.update({ "/data": {
        #                 "bind": "/data",
        #                 "mode": "rw"
        #             }})
        entrypoint = None
        if job.run_id.startswith("job-"):
            entrypoint="bash"
        try:
            container: Container = self.client.containers.run(
                image=find_container.image,
                name=job.run_id,
                user= docker_uid,
                group_add=["users",str(sock_gid)],
                command=command,
                network=network,
                entrypoint=entrypoint,
                volumes={
                    job.output_dir: {
                        "bind": job.output_dir,
                        "mode": "rw"
                    },
                    script_dir:{
                        "bind": script_dir,
                        "mode": "rw"
                    },
                    work_dir: {
                        "bind": work_dir,
                        "mode": "rw"
                    },
                    pipeline_dir: {
                        "bind": pipeline_dir,
                        "mode": "rw"
                    },
                    base_dir: {
                        "bind": base_dir,
                        "mode": "rw"
                    },
                    **volumes,
                    "/usr/bin/docker":{
                        "bind": "/usr/bin/docker",
                        "mode": "rw"
                    },
                    "/tmp/brave.sock": {
                        "bind": "/tmp/brave.sock",
                        "mode": "rw"
                    },
                    "/var/run/docker.sock": {
                        "bind": "/var/run/docker.sock",
                        "mode": "rw"
                    }
                },
                environment={
                    **envionment,
                    "PIPELINE_DIR": f"{pipeline_dir}",
                    "OUTPUT_DIR":f"{job.output_dir}",
                    "COMMON_SCRIPT_DIR":f"{connom_script_dir}/common"
                },
                working_dir=job.output_dir,
                detach=True,
                labels={
                    "job_id": job.run_id,
                    "analysis_id": job.analysis_id,
                    "project": "brave",
                    "user": str(user_id),
                    "type":job.run_id.split("-")[0],
                    **labels
                },
                 ports=port
                # remove=True
            )
            
        except Exception as e:
            print(f"Error running container {job.run_id}: {e}")
            self.to_remove.append(job.run_id)
            raise e
        if container.id is None:
            raise RuntimeError("Container did not return a valid ID")
        
        self.containers[job.run_id] = container
        container.reload()
        ports = container.attrs['NetworkSettings']['Ports']
        job.ports = ports
        # if self._monitor_task is None:
        #     self._monitor_task = asyncio.create_task(self._monitor_containers())
        self._list_running()
        asyncio.run(self.event_bus.dispatch(
            RoutersName.ANALYSIS_EXECUTER_ROUTER,
            AnalysisExecutorEvent.ON_ANALYSIS_STARTED,
            job
        ) )  
        return container.id



    async def _monitor_containers(self):
        while True:
            try:
                for job_id, container in list(self.containers.items()):
                    try:
                        container.reload()
                        run_id = AnalysisId(run_id=job_id)
                        if container.status in ("exited", "dead"):
                            exit_code = container.attrs["State"]["ExitCode"]

                            if exit_code == 0:
                                # 成功退出，自动删除容器
                                print(f"[{job_id}] 执行成功，删除容器")
                                container.remove(force=True)
                                self.containers.pop(job_id, None)
                                
                                await self.event_bus.dispatch(
                                    RoutersName.ANALYSIS_EXECUTER_ROUTER,
                                    AnalysisExecutorEvent.ON_ANALYSIS_COMPLETE,
                                    run_id
                                )
                                self._list_running()
                            else:
                                # 执行失败，保留容器调试
                                print(f"[{job_id}] 执行失败（ExitCode={exit_code}），保留容器")
                                self.containers.pop(job_id, None)  # 不删除容器，仅移出监控
                                await self.event_bus.dispatch(
                                    RoutersName.ANALYSIS_EXECUTER_ROUTER,
                                    AnalysisExecutorEvent.ON_ANALYSIS_FAILED,
                                    run_id
                                )  
                                self._list_running()
                    except Exception as e:
                        print(f"Error monitoring container {job_id}: {e}")
                        self.to_remove.append(job_id)
                        self._list_running()
                

                for job_id in self.to_remove:
                    if job_id in self.containers:
                        self.containers.pop(job_id, None)
                    run_id = AnalysisId(run_id=job_id)
                    await self.event_bus.dispatch(
                            RoutersName.ANALYSIS_EXECUTER_ROUTER,
                            AnalysisExecutorEvent.ON_ANALYSIS_COMPLETE,
                            run_id
                        )
                    self.to_remove.remove(job_id)
                await asyncio.sleep(self._monitor_interval)
            except Exception as e:
            
                print(f"Error auto removing container {job_id}: {e}")
                traceback.print_exc()
                pass

    def get_logs(self, job_id: str) -> str:
        try:
            logs = self.client.containers.get(job_id).logs()
            print(f"logs: {logs}")
            if logs is None:
                return ""
            return logs.decode()
        except Exception as e:
            print(f"Error getting logs for container {job_id}: {e}")
            return ""

    def stop_job(self, job_id: str) -> None:
        try:
            self.client.containers.get(job_id).stop()
            self._list_running()
        except Exception as e:
            print(f"Error stopping container {job_id}: {e}")
            pass
    
    async def remove_job(self, job_id: str) -> None:
        try:
            self.client.containers.get(job_id).remove(force=True)
            self._list_running()
        except Exception as e:
            print(f"Error removing container {job_id}: {e}")
            pass
    
    def get_image_name(self,container):
        tags = container.image.tags
        if tags and len(tags)>0:
            return tags[0]
        else:
            return "<none>:<none>"

        
    def _list_running(self,recover=False) :
        label_filter = {"label": "project=brave"}
        # 获取运行中的容器（filtered by label）
        containers = self.client.containers.list(filters=label_filter)
        if recover:
            self.recover_running_containers(containers)
        containers = [{"image":self.get_image_name(container),"run_id":container.name,"name":container.name,"id":container.id} for container in containers]
        self.running_conntainer = containers
        
        return containers
    
    async def refresh_list_running(self) :
        return self._list_running()

    
    async def list_running(self) :
        return self.running_conntainer

    async def update_images_status(self):
        with get_engine().begin() as conn: 
            container_list = container_service.list_container(conn)
            for item in container_list:
                image = self.get_image(item.image)
                if image:
                    container_service.update_container(conn,
                                                       item.container_id,
                                                       {"image_id":image.id,"image_status":"exist"})
                else:
                    container_service.update_container(conn,
                                                       item.container_id,
                                                       {"image_status":"not_exist"})
             
            

    def get_image(self, image_name):
        try:
            image = self.client.images.get(image_name)
            return image
        except ImageNotFound:
            print(f"Image ({image_name}) not exist!")
            return None
        except Exception as e:
            return None
    
    def pull_image_with_log(self,container_id,image_name: str):
        client = docker.APIClient(base_url="unix://var/run/docker.sock")
        stream = client.pull(image_name.strip(), stream=True, decode=True)
        log_file = f"{self.setting.BASE_DIR}/log"
        if not  os.path.exists(log_file):    
            os.makedirs(log_file) 
        print(f"pull {image_name} log :{log_file}/{container_id}.log")
        with open(f"{log_file}/{container_id}.log", "w", encoding="utf-8") as f:
            for log in stream:
                log_str = json.dumps(log, ensure_ascii=False)
                # print(log_str)  # 继续打印到控制台
                f.write(log_str + "\n")  # 写入文件

        # for log in stream:
            
        #     print(json.dumps(log, indent=2, ensure_ascii=False))

    async def pull_image(self,container_id,image_name):
        print(f"pull {image_name}")
        # a = self.client.images.pull(image_name)
        
        await asyncio.to_thread(self.pull_image_with_log,container_id,image_name)
        image = self.get_image(image_name)
        if image:
            print(f"pull {image_name} complete!")
            with get_engine().begin() as conn: 
                container_service.update_container(conn,
                                                    container_id,
                                                    {"image_id":image.id,"image_status":"exist"})
        run_id = AnalysisId(run_id=container_id)
        # asyncio.create_task() 
        await self.event_bus.dispatch(
            RoutersName.ANALYSIS_EXECUTER_ROUTER,
            AnalysisExecutorEvent.ON_CONTAINER_PULLED,
            run_id
        )
    async def get_container_attr(self,container_id):
        try:
            container = self.client.containers.get(container_id)
            container.reload()
            return container.attrs
        except Exception as e:
            print(f"Error getting attrs for container {container_id}: {e}")
            return None
    async def get_image_attr(self,image_name):
        try:
            image = self.client.images.get(image_name)
            return image.attrs
        except Exception as e:
            print(f"Error getting attrs for image {image_name}: {e}")
            return None


import asyncio
import aiohttp
from Bio import Entrez
import xml.etree.ElementTree as ET
from xml.etree import ElementTree

async def get_pmc_fulltext(pmcid):
    url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"
    params = {
        "db": "pmc",
        "id": pmcid,
        "rettype": "full",
        "retmode": "xml",
        "api_key": "b4761f644bab9143a8bbef4c1d124227d208"
    }
    async with aiohttp.ClientSession() as session:
        async with session.get(url, params=params) as resp:
            print(f"查询全文: {pmcid}")
            text = await resp.text()

            try:
                root = ElementTree.fromstring(text)
                # 提取<body>内所有<p>内容，和之前的同步例子类似
                ns = ''
                if root.tag.startswith('{'):
                    ns = root.tag.split('}')[0] + '}'

                body = root.find(f".//{ns}body")
                if body is None:
                    print("未找到<body>节点，返回空全文")
                    fulltext = ""
                else:
                    paragraphs = []
                    def recursive_extract(element):
                        for child in element:
                            if child.tag == f"{ns}p":
                                paragraphs.append(''.join(child.itertext()).strip())
                            else:
                                recursive_extract(child)
                    recursive_extract(body)
                    fulltext = "\n\n".join(paragraphs)

                return fulltext
            except Exception as e:
                print(f"解析全文XML出错，返回内容:\n{text}")
                raise e


import json
import os
import textwrap
from typing import Any, Optional
from .base_analysis import BaseAnalysis
import  brave.api.service.pipeline as pipeline_service
from brave.api.core.evenet_bus import EventBus

class NextflowAnalysis(BaseAnalysis):
    def __init__(self, event_bus:EventBus) -> None:
        super().__init__(event_bus)



    
    

    def _get_command(self,analysis_id,output_dir,cache_dir,params_path,work_dir,executor_log,component_script,trace_file,workflow_log_file,pieline_dir_with_namespace,script_type) -> str:
        nextflow_config =  f"{pieline_dir_with_namespace}/nextflow.config"
        if  not os.path.exists(nextflow_config):
            # config_arg = f" -c {nextflow_config}"
            with open(nextflow_config,"w") as f:
                f.write("")
        command =  textwrap.dedent(f"""
            export BRAVE_WORKFLOW_ID={analysis_id}
            export NXF_CACHE_DIR={cache_dir}
            nextflow -log {executor_log} run -offline -resume  \\
                -ansi-log false \\
                {component_script} \\
                -params-file {params_path} \\
                -w {work_dir} \\
                -c {nextflow_config} \\
                -with-trace {trace_file} | tee {workflow_log_file} ; exit ${{PIPESTATUS[0]}}
            """)
        # -plugins nf-hello@0.7.0 \\

        return command
        
    def write_config(self,output_dir,component_script):
        script_config_file = f"{output_dir}/nextflow.config"
        script_config =  textwrap.dedent(f"""
        trace.overwrite = true
        docker{{
            enabled = true
            runOptions = '--user $(id -u):$(id -g)'
        }}
        """)
        with open(script_config_file, "w") as f:
            f.write(script_config)
        return script_config_file
  


#!/usr/bin/env python
# coding: utf-8

# In[56]:


import pandas as pd
from brave.api.config.db import get_engine
from brave.microbe.models.core import t_taxonomy

def read_nodes(file):
    """读取 nodes.dmp"""
    cols = [
        "tax_id", "parent_tax_id", "rank", "embl_code", "division_id",
        "inherited_div_flag", "genetic_code_id", "inherited_gc_flag",
        "mitochondrial_genetic_code_id", "inherited_mgc_flag",
        "genbank_hidden_flag", "hidden_subtree_root_flag", "comments"
    ]
    df = pd.read_csv(
        file, sep="|",header=None, usecols=range(len(cols)),
        names=cols
    )
    df = df.applymap(lambda x: x.strip() if isinstance(x, str) else x)
    df = df[["tax_id","parent_tax_id","rank","division_id"]]
    return df


# In[58]:


# read_nodes("/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/nodes.dmp.head")


# In[25]:


def read_names(file):
    """读取 names.dmp，只保留 scientific name"""
    cols = ["tax_id", "name_txt", "unique_name", "name_class"]
    df = pd.read_csv(
        file, sep="|", header=None, usecols=range(len(cols)),
        names=cols
    )
    df = df.applymap(lambda x: x.strip() if isinstance(x, str) else x)
    df = df[df["name_class"] == "scientific name"]
    df = df[["tax_id", "name_txt"]].rename(columns={"name_txt": "scientific_name"})
    return df 


# In[46]:


# df = read_names("/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/names.dmp")


# In[40]:


def read_division(file):
    """读取 division.dmp"""
    cols = ["division_id", "division_cde", "division_name", "comments"]
    df = pd.read_csv(
        file, sep="|", engine="python", header=None, usecols=range(len(cols)),
        names=cols
    )
    df = df.applymap(lambda x: x.strip() if isinstance(x, str) else x)
    df = df[["division_id","division_name"]]
    return df


# In[45]:


# read_division("/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/division.dmp")


# In[41]:


def merge_all(nodes_file, names_file, division_file):
    """合并三张表"""
    nodes = read_nodes(nodes_file)
    names = read_names(names_file)
    division = read_division(division_file)

    # 合并 nodes + names
    merged = pd.merge(nodes, names, on="tax_id", how="left")

    # 合并 division
    merged = pd.merge(merged, division, on="division_id", how="left")

    return merged


# In[59]:




# In[48]:


# len(set(df["scientific_name"]))


# In[49]:


# len(df["scientific_name"])


# In[52]:


# len(set(df["tax_id"]))


# In[53]:


# len(df["tax_id"])



if __name__ == "__main__":
    # nodes_file = "nodes.dmp"
    # names_file = "names.dmp"
    # division_file = "division.dmp"

    # taxonomy = merge_all(nodes_file, names_file, division_file)
    
    taxonomy = merge_all("/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/nodes.dmp","/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/names.dmp","/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/division.dmp")

    # 保存到 CSV
    # taxonomy.to_csv("taxonomy.csv", index=False)
    with get_engine().begin() as conn:
        stmt = t_taxonomy.insert().values(taxonomy.to_dict(orient="records"))
        conn.execute(stmt)

    # 或者保存到 SQLite / MySQL (示例 SQLite)
    # import sqlite3
    # conn = sqlite3.connect("taxonomy.db")
    # taxonomy.to_sql("taxonomy", conn, if_exists="replace", index=False)
    # conn.close()

    # print("整合完成，共有记录：", len(taxonomy))



from typing import Optional,Any
from pydantic import BaseModel

class AnalysisInput(BaseModel):
    id: Optional[int]= None
    project: Optional[str]
    samples: list
    analysis_method:str
    analysis_name:str
    
    # analysis_name: Optional[str]
    # work_dir: Optional[str]
    # output_dir: Optional[str]
class QueryAnalysis(BaseModel):
    analysis_id: Optional[str]=None
    analysis_method: Optional[str]=None
    component_id: Optional[str]=None
    component_ids: Optional[list[str]] =None
    is_report: Optional[bool] =None
    project: Optional[str]=None

class Analysis(BaseModel):
    id: Optional[int]
    project: Optional[str]
    analysis_id:str
    component_id: Optional[str]
    analysis_method: Optional[str]
    analysis_name: Optional[str]
    input_file: Optional[str]
    request_param: Optional[str]
    work_dir: Optional[str]
    output_dir: Optional[str]
    params_path: Optional[str]
    output_format: Optional[str]
    command_path: Optional[str]
    pipeline_script: Optional[str]
    parse_analysis_module: Optional[str]
    process_id: Optional[str]
    # analysis_status: Optional[str]


class AnalysisExecuterModal(BaseModel):
    id: Optional[int]=None
    project: Optional[str]=None
    analysis_id: str
    run_id:str
    component_id: Optional[str]=None
    analysis_method: Optional[str]=None
    analysis_name: Optional[str]=None
    input_file: Optional[str]=None
    request_param: Optional[str]=None
    work_dir: Optional[str]=None
    output_dir: Optional[str]=None
    params_path: Optional[str]=None
    output_format: Optional[str]=None
    command_path: Optional[str]=None
    pipeline_script: Optional[str]=None
    parse_analysis_module: Optional[str]=None
    process_id: Optional[str]=None
    analysis_status: Optional[str]=None
    command_log_path: Optional[str]=None
    run_type:Optional[str]=None
    image: Optional[str]=None
    container_id: Optional[str]=None
    # change_uid: Optional[str]=None
    ports: Optional[Any]=None

class AnalysisId(BaseModel):
    run_id: str
    run_type:Optional[str]=None


class UpdateProject(BaseModel):
    project: Optional[list[str]] = None
    analysis_id: Optional[str]=None



# container/slurm_backend.py
import subprocess

from brave.api.executor.models import JobSpec
from .base import JobExecutor
from brave.api.core.evenet_bus import EventBus  
class SlurmExecutor(JobExecutor):
    def __init__(self,event_bus:EventBus):
        self.event_bus = event_bus
    async def _do_submit_job(self, job_spec):
        pass
        # script_path = job_spec["script_path"]
        # result = subprocess.run(["sbatch", script_path], capture_output=True, text=True)
        # job_id = result.stdout.strip().split()[-1]
        # return job_id

    def get_logs(self, job_id: str) -> str:
        log_path = f"/path/to/logs/{job_id}.out"
        with open(log_path) as f:
            return f.read()

    def stop_job(self, job_id: str) -> None:
        subprocess.run(["scancel", job_id])



import glob
import os
import json 
from brave.api.utils.from_glob_get_file import from_glob_get_file
# "dir_path":dir_path,
# "analysis": result,
# "output_file":item['output_file'],
# "args":moduleArgs

# def get_content(file,file_list,suffix,replace_map):
#     content_dict = {}
   
#     if suffix:
#         # file_list = glob.glob(f"{file}/*")
#         for item in file_list:
#             analysis_name = get_analysis_name(item,suffix)
#             filename = os.path.basename(item).replace(f"{analysis_name}.","").replace(analysis_name,"")
#             if filename in replace_map:
#                 filename = replace_map[filename]
#             content_dict[filename] = item
#     else:
#         # file_list = glob.glob(f"{file}/*/*")
#         for item in file_list:
#             analysis_name = get_analysis_name(item,suffix)
#             filename = os.path.basename(item).replace(analysis_name,"")
#             content_dict[filename] = item
#     return json.dumps(content_dict)

        

# def get_analysis_name(file,suffix):
#     if suffix:
#         return os.path.basename(file).replace(suffix,"")
#     else:
#         return os.path.dirname(file)

def parse(dir_path,file_format,sample_list):

    form_data = from_glob_get_file(file_format, dir_path)
    sample_list_dict = {item['sample_name']:item for item in sample_list}
    #  analysis_key,software,content_type,content
    result_data = [
        {
            "file_name": item['sample_name'],
            "sample_source": sample_list_dict.get(item['sample_name'], {}).get('sample_source'),
            # "sample_name": item['sample_name'], # analysis result 表中没有 sample_name 字段
            "content_type": "json",
            "content":json.dumps({k:v for k,v in item.items() if k != 'sample_name'   }),
        }
        for item in form_data   ]
    return result_data


    # if not replace_map:
    #     replace_map = {}

    # file_list = glob.glob(f"{dir_path}/{pattern}")
    # result_data = []
    # for file in file_list:
    #     analysis_name = get_analysis_name(file,suffix)
    #     content = get_content(file,file_list,suffix,replace_map)
    #     result_data.append((analysis_name, name,"json",content))


    # return result_data





# def get_content(file):
#     return json.dumps({
#         "bam":file,
#         "log":file.replace(".bam","")+".bowtie2.log"
#     })

# def parse(dir_path,analysis):
#     file_list = glob.glob(f"{dir_path}/*/*.bam")
#     result_data = [(os.path.basename(file).replace(".bam",""),"bowtie2","json",get_content(file)) for file in file_list]
#     return result_data



# def get_json(file):
#     return json.dumps({
#         "fastq1":file,
#         "fastq2":file.replace("_host_removed_R1.fastq.gz","_host_removed_R2.fastq.gz")
#     })

# def parse(dir_path,analysis):
#     file_list = glob.glob(f"{dir_path}/*_host_removed_R1.fastq.gz")
#     result_data = [(os.path.basename(file).replace("_host_removed_R1.fastq.gz",""),"samtools","json",get_json(file)) for file in file_list]
#     return result_data




from brave.api.service.components.base import ComponentHandler
from brave.api.service.components.shared_handler import SharedComponentHandler
from brave.api.service.components.independent_handler import IndependentComponentHandler


class ComponentHandlerFactory:
    @staticmethod
    def get_handler(component_model: str) -> ComponentHandler:
        if component_model == "shared":
            return SharedComponentHandler()
        elif component_model == "independent":
            return IndependentComponentHandler()
        else:
            raise ValueError(f"未知组件模型: {component_model}")



from operator import and_, or_
from brave.api.config.db import get_engine
from brave.api.models.core import analysis as t_analysis, t_pipeline_components,t_project
from sqlalchemy import select,update
from fastapi import HTTPException
import json
from brave.api.schemas.analysis import QueryAnalysis
import  brave.api.service.pipeline as pipeline_service
import importlib
import hashlib
import os
from brave.api.enum.component_script import ScriptName
from brave.api.models.core import t_container
from sqlalchemy.orm import aliased

def get_parse_analysis_result_params(conn,analysis_id):
    stmt = select(t_analysis).where(t_analysis.c.analysis_id == analysis_id)
    result = conn.execute(stmt).mappings().first()
    if not result:
        raise HTTPException(status_code=404, detail=f"Analysis with id {analysis_id} not found")
    component_id = result['component_id']
    component_ = pipeline_service.find_pipeline_by_id(conn, component_id)
    if not component_:
        raise HTTPException(status_code=404, detail=f"Component with id {component_id} not found")
  
    file_format_list = []
    if component_["component_type"] == "pipeline":
        component_ = pipeline_service.get_pipeline_v2(conn,component_id)
        softwareList = component_['software']
        component_file_list = [outputFile for item in softwareList if 'outputFile' in item for outputFile in item['outputFile']]
        file_format_list = [
            {"dir":item['dir'],"fileFormat":item['fileFormat'],"name":item['name'],"component_id":item['component_id']}
            for item in component_file_list if 'fileFormat' in item
        ]
        pass
    else:
     
        component_file_list = pipeline_service.find_component_by_parent_id(conn,component_id,"software_output_file")
        component_file_content_list = [{**json.loads(item.content),"component_id":item['component_id']} for item in component_file_list]
        file_format_list = [
            {"dir":item['dir'],"fileFormat":item['fileFormat'],"name":item['name'],"component_id":item['component_id']}
            for item in component_file_content_list if 'fileFormat' in item
        ]
        # component_file_list = []
    # component_file_list = pipeline_service.find_component_by_parent_id(conn,component_id,"software_output_file")
    component_ = {
            **{k:v for k,v in component_.items() if k != "content"},
            **json.loads(component_['content'])
        }
    # try:
    #     component_content = json.loads(component_.content)
    # except Exception as e:
    #     raise HTTPException(status_code=500, detail=f"Failed to parse component content: {e}")
    # parse_analysis_result_module = component_content.get('parseAnalysisResultModule')
    
    
    
    # if len(component_file_list) == 0:
    #     return {"error":"组件没有添加输出文件,请检查!"}
        # raise HTTPException(status_code=500, detail=f"组件{component_id}没有添加输出文件,请检查!")
    
    # if not file_format_list:
    #     return {"error":"组件的输出文件没有配置fileFormat!请检查!"}
        # raise HTTPException(status_code=500, detail=f"组件{component_id}的输出文件没有配置fileFormat!请检查!")


    module_info = pipeline_service.find_component_module(component_, ScriptName.output_parse)
    if not module_info:
        raise HTTPException(status_code=500, detail=f"组件{component_id}的输出解析模块没有找到!")

    py_module = module_info['module']
    module = importlib.import_module(py_module)
    parse = getattr(module, "parse")

    return {
        "analysis":result,
        "file_format_list":file_format_list,
        "parse":parse
    }



def execute_parse(analysis,parse,file_format_list):
    analysis_params_path = analysis.get('params_path')
    if not analysis_params_path or not os.path.exists(analysis_params_path):
        raise HTTPException(status_code=500, detail=f"Analysis params_path {analysis_params_path} not found")
    
    with open(analysis_params_path,"r") as f:
        analysis_params = json.load(f)
        if "sample_list" in analysis_params:
            sample_list = analysis_params["sample_list"]
            



    result_dict = {}
    result_list = []
    for item in file_format_list:        
        dir_path = f"{analysis['output_dir']}/output/{item['dir']}"
        res = None    
        args = {
            "dir_path":dir_path,
            # "analysis": dict(result),
            "file_format":item['fileFormat'],
            "sample_list":sample_list
            # "args":moduleArgs,
        
        }
        res = parse(**args)
       
        for sub_item in  res:
            sub_item.update({
                "component_id":item['component_id'],
                # "analysis_name":item['name'],
                # "analysis_method":item['name'],
                "project":analysis['project'],
                "analysis_id":analysis['analysis_id'],
                "analysis_type":"upstream_analysis",
                "analysis_result_hash":hashlib.md5(sub_item['content'].encode()).hexdigest()
                })
        result_dict.update({item['name']:res})
        result_list = result_list + res
    return result_list,result_dict

def add_run_id(item):
    item= dict(item)
    if item['job_status'] == "running":
        item['run_id'] = f"job-{item['analysis_id']}"
        item['run_type'] = "job"
    elif item['server_status'] == "running":
        item['run_id'] = f"server-{item['analysis_id']}"
        item['run_type'] = "server"
    return item

def find_running_analysis(conn):
    stmt = select(t_analysis).where(or_(t_analysis.c.job_status == "running",t_analysis.c.server_status == "running"))
    result = conn.execute(stmt).mappings().all()
    result = [add_run_id(item) for item in result]
    return result

def get_all_files_recursive(directory,dir_name,file_dict):
    file_list=[]
    for root, dirs, files in os.walk(directory):
        for file in files:
            file_list.append(os.path.join(root, file).replace(directory,""))
    return file_dict.update({dir_name:file_list})

def get_file_dict(file_format_list,output_dir):
    file_dict={}
    for item in file_format_list:
        dir_path = f"{output_dir}/output/{item['dir']}"
        get_all_files_recursive(dir_path,item['dir'],file_dict)
    return file_dict

def find_analysis_by_id(conn,analysis_id):
    stmt = select(t_analysis).where(t_analysis.c.analysis_id == analysis_id)
    result = conn.execute(stmt).mappings().first()
    return result


async def finished_analysis(analysis_id,run_type,status):
    with get_engine().begin() as conn:  
        if run_type == "job":
            stmt = (
                update(t_analysis)
                .where(t_analysis.c.analysis_id == analysis_id)
                .values(job_status = status)
            )
        elif run_type == "server":
            stmt = (
                update(t_analysis)
                .where(t_analysis.c.analysis_id == analysis_id)
                .values(server_status = status)
            )
        else:
            raise ValueError(f"Invalid run_type: {run_type}")
        
        conn.execute(stmt)
        conn.commit()
    print(f"Analysis {analysis_id} {status}")

async def update_ports(analysis_id,ports):
    with get_engine().begin() as conn:  
        stmt = (
            update(t_analysis)
            .where(t_analysis.c.analysis_id == analysis_id)
            .values(ports = ports)
        )
        conn.execute(stmt)
    print(f"Analysis {analysis_id} {ports}")

def update_report(conn,analysis_id,is_report):
    stmt = (
        update(t_analysis)
        .where(t_analysis.c.analysis_id == analysis_id)
        .values(is_report = is_report)
    )
    conn.execute(stmt)

async def update_url(analysis_id,url):
    with get_engine().begin() as conn:  
        stmt = (
            update(t_analysis)
            .where(t_analysis.c.analysis_id == analysis_id)
            .values(url = url)
        )
        conn.execute(stmt)
    print(f"update Analysis {analysis_id} {url}")


def list_analysis(conn,query:QueryAnalysis):
    conditions = []
    if query.analysis_id:
        conditions.append(t_analysis.c.analysis_id == query.analysis_id)
    if query.analysis_method:
        conditions.append(t_analysis.c.analysis_method == query.analysis_method)
    if query.component_id:
        conditions.append(t_analysis.c.component_id == query.component_id)
    if query.project:
        conditions.append(t_analysis.c.project == query.project)
    if query.component_ids:
        conditions.append(t_analysis.c.component_id.in_(query.component_ids))
    if query.is_report:
        conditions.append(t_analysis.c.is_report)
    # t_sub_container = aliased(t_container)

    stmt = select(
        t_analysis,
        t_pipeline_components.c.component_name.label("component_name"),
        t_pipeline_components.c.order_index.label("component_order_index"),
        # t_pipeline_components.c.label.label("component_label"),
        t_pipeline_components.c.component_type.label("component_type"),
        t_project.c.project_name.label("project_name"),
        t_container.c.name.label("container_name"),
        t_container.c.image.label("container_image"),
        t_container.c.container_id.label("container_id"),
        t_container.c.image_status.label("image_status"),
        t_container.c.image_id.label("image_id")
        # t_sub_container.c.name.label("sub_container_name"),
        # t_sub_container.c.image.label("sub_container_image")
    )

    stmt = stmt.select_from(
        t_analysis.outerjoin(t_pipeline_components,t_analysis.c.component_id==t_pipeline_components.c.component_id)
        .outerjoin(t_project,t_analysis.c.project==t_project.c.project_id)
        .outerjoin(t_container,t_pipeline_components.c.container_id==t_container.c.container_id)
        # .outerjoin(t_sub_container,t_pipeline_components.c.sub_container_id==t_sub_container.c.container_id)
        )
    if conditions:
        stmt = stmt.where(and_(*conditions) if len(conditions) > 1 else conditions[0])
    return conn.execute(stmt).mappings().all()


def update_extra_project(conn,analysis_id,project):
    stmt = (
        update(t_analysis)
        .where(t_analysis.c.analysis_id == analysis_id)
        .values(extra_project_ids = json.dumps(project))
    )
    conn.execute(stmt)


from brave.api.models.core import samples

def find_by_sample_name(conn,sample_name):
    stmt = samples.select().where(samples.c.sample_name==sample_name)
    result = conn.execute(stmt).mappings().first()
    return result

def find_by_sample_name_and_project(conn,sample_name,project):
    stmt = samples.select().where(samples.c.sample_name==sample_name,samples.c.project==project)
    result = conn.execute(stmt).mappings().first()
    return result
    
def add_sample(conn,sample_data:dict):
    stmt = samples.insert().values(sample_data)
    conn.execute(stmt)


def find_by_sample_name_list(conn,sample_name_list):
    stmt = samples.select().where(samples.c.sample_name.in_(sample_name_list))
    result = conn.execute(stmt).mappings().all()
    return result




import json
from fastapi import APIRouter
from brave.api.schemas.project import AddProject,UpdateProject
from brave.api.service import project_service
from brave.api.config.db import get_engine

project_api = APIRouter(prefix="/project")

@project_api.post("/add-project")
async def add_project(AddProject:AddProject):
    with get_engine().begin() as conn:
        project_id = await project_service.add_project(conn,AddProject)
        return {"project_id":project_id}

@project_api.post("/update-project")
async def update_project(UpdateProject:UpdateProject):
    with get_engine().begin() as conn:
        project_id = await project_service.update_project(conn,UpdateProject)
        return {"project_id":project_id}
    
def get_one_project(item):
    if not item:
        return {}
    item = dict(item)
    try:
        item['metadata_form'] = json.loads(item['metadata_form'])
    except:
        item["metadata_form"] = []

    return item

@project_api.get("/list-project")
async def list_project():
    with get_engine().begin() as conn:
        projects = await project_service.list_project(conn)
        projects = [get_one_project(item) for item in projects]
        return projects


@project_api.get("/find-by-project-id/{project_id}")
async def find_by_project_id(project_id:str):
    with get_engine().begin() as conn:
        project = await project_service.find_by_project_id(conn,project_id)
        return get_one_project(project)

@project_api.delete("/delete-project/{project_id}")
async def delete_project(project_id:str):
    with get_engine().begin() as conn:
        project_id = await project_service.delete_project(conn,project_id)
        return {"project_id":project_id}


from abc import ABC, abstractmethod
from brave.api.core.event import WorkflowEvent
from brave.api.executor.models import JobSpec
from brave.api.core.routers.workflow_event_router import WorkflowEventRouter
from brave.api.schemas.analysis import AnalysisExecuterModal

class JobExecutor(ABC):
    def __init__(self):
        pass
    async def submit_job(self, job_spec: AnalysisExecuterModal) -> str:
        if self.is_already_running(job_spec.run_id):
            await self.remove_job(job_spec.run_id)
            # raise Exception(f"Job {job_spec.job_id} is already running")
        await self._do_submit_job(job_spec)
        # await self.router.dispatch(WorkflowEvent.ON_JOB_SUBMITTED,{"event": "on_job_submitted", "job_id": job_spec.job_id})
        return job_spec.analysis_id



    @abstractmethod
    async def _do_submit_job(self, job_spec: AnalysisExecuterModal) -> str:
        pass

    @abstractmethod
    def get_logs(self, job_id: str) -> str:
        pass

    @abstractmethod
    def stop_job(self, job_id: str) -> None:
        pass
    
    def is_already_running(self, job_id: str) -> bool:
        return False 
    
    async def remove_job(self, job_id: str) -> None:
        pass

    async def refresh_list_running(self):
        pass

    async def list_running(self):
        return []

    def get_image(self,image_name):
       pass
    async def pull_image(self,container_id,image_name):
        pass

    async def get_container_attr(self,container_id):
        pass
    async def get_image_attr(self,image_name):
        pass

    async def update_images_status(self):
        pass


from ast import Dict, Set
import asyncio
import psutil
from watchfiles import awatch
from datetime import datetime
from brave.api.config.db import get_engine
from brave.api.models.core import analysis
from sqlalchemy import select, update
import logging
from importlib.resources import files
from importlib import import_module
import inspect
from brave.api.service.sse_service import SSESessionService
from brave.api.service.analysis_result_parse import AnalysisResultParse
from collections import defaultdict
from typing import Dict, Set
from brave.api.service.listener_files_service import ListenerFilesService
# 创建 logger
logger = logging.getLogger(__name__)

class ProcessMonitor:
    def __init__(self, 
    sse_service: SSESessionService,
    analysis_result_parse_service: AnalysisResultParse,
    listener_files_service: ListenerFilesService,
    check_interval: int = 5,
     
    ):
        self.queue_process = asyncio.Queue()
        self.queue_lock = asyncio.Lock()  # 保证数据库更新和队列操作安全
        self.check_interval = check_interval  # 检查间隔
        # self.listener_files = self._load_listener_files()
        self.sse_service = sse_service
        self.analysis_id_to_queue: Dict[str, Set[asyncio.Queue]] = defaultdict(set)
        self.analysis_result_parse_service = analysis_result_parse_service
        self.listener_files_service = listener_files_service
    # def _load_listener_files(self):
    #     """
    #     加载监听器文件列表
    #     """
    #     listener_files = files("brave.api.listener")
    #     return [
    #         item.stem
    #         for item in listener_files.iterdir()
    #         if item.is_file() and item.name.endswith(".py") and item.name != "__init__.py" and item.name.startswith("file")
    #     ]

    # async def execute_listener(self, func, args):
    #     """
    #     执行监听器的回调函数
    #     """
    #     if isinstance(self.listener_files, list) and len(self.listener_files) > 0:
    #         for name in self.listener_files:
    #             full_module = f"brave.api.listener.{name}"
    #             mod = import_module(full_module)
    #             if hasattr(mod, func):
    #                 run_func = getattr(mod, func)
    #                 if inspect.iscoroutinefunction(run_func):
    #                     asyncio.create_task(run_func(**args))
    #                 else:
    #                     await asyncio.to_thread(run_func, **args)

    async def startup_process_event(self):
        """
        初始化队列中的任务，加载数据库中的数据到队列中
        """
        with get_engine().begin() as conn:
            stmt = select(analysis).where(analysis.c.process_id != None)
            results = conn.execute(stmt).all()
            for row in results:
                item = dict(row._mapping)
                await self.queue_process.put(item)

        logger.info(f"队列初始化完毕，任务数：{self.queue_process.qsize()}")
        # 启动进程检查工作
        # asyncio.create_task()
        await self.check_process_worker()

    async def check_process_worker(self):
        """
        负责检查进程是否存在，每隔一段时间重新检查队列中的任务
        """
        while True:
            item = await self.queue_process.get()
            process_id = item.get("process_id")
            analysis_id = item.get("analysis_id")
            logger.info(f"检查分析任务 id={analysis_id}, pid={process_id}")

            try:
                pid_int = int(process_id)
                proc = psutil.Process(pid_int)
                proc_name = proc.name().lower()
                if "bash" not in proc_name:
                    raise psutil.NoSuchProcess(f"进程 {process_id} 不是 nextflow")

            except (psutil.NoSuchProcess, ValueError) as e:
                logger.warning(f"进程 {process_id} 不存在或非 nextflow，清理数据库: {e}")
                # 更新数据库，将 process_id 设为 None
                await self.clean_up_process(item)
            else:
                # 进程存在且符合要求，延迟后重新入队
                await asyncio.sleep(self.check_interval)
                await self.queue_process.put(item)
            finally:
                self.queue_process.task_done()

    async def clean_up_process(self, item):
        """
        清理数据库中的 process_id 字段，并触发监听器事件
        """
        analysis_id = item.get("analysis_id")
        # 确保数据库操作的线程安全
        async with self.queue_lock:
            await self.listener_files_service.execute_listener("process_end", {
                "analysis": item,
                "sse_service": self.sse_service,
                "analysis_result_parse_service":self.analysis_result_parse_service})

    async def add_process(self, analysis_dict):
        async with self.queue_lock:
            await self.queue_process.put(analysis_dict)

    async def get_process(self):
        return await self.queue_process.get()


# service.py

import asyncio
import threading
from collections import defaultdict
from functools import lru_cache
from typing import Dict, Set
import json
from fastapi import Request
from fastapi.responses import StreamingResponse


class SSEService:
    def __init__(self):
        self.global_queue = asyncio.Queue()
        self.connected_clients = set()

    async def broadcast_loop(self):
        # current_loop = asyncio.get_event_loop()
        # print(f"broadcast_loop 事件循环：{current_loop}")
        while True:
            msg = await self.global_queue.get()
            print(f"广播消息: {msg} 客户端数量: {len(self.connected_clients)}")
            for q in self.connected_clients.copy():
                await q.put(msg)

    async def event_generator(self, request, client_queue):
        try:
            # 首条立即推送，确保客户端立即建立连接
            # yield "data: connected\n\n"
            while True:
                if await request.is_disconnected():
                    print("请求关闭!")
                    break
                try:
                    msg = await asyncio.wait_for(client_queue.get(), timeout=10)
                    print(f"产生消息: {msg}!")
                    yield f"data: {msg}\n\n"
                except asyncio.TimeoutError:
                    yield ": keep-alive\n\n"
        except asyncio.CancelledError:
            print("连接被取消")
        finally:
            print("finally请求关闭!")
            self.connected_clients.discard(client_queue)

    async def push_message(self, msg: str):
        await self.global_queue.put(msg)

    def add_client(self, client_queue):
        self.connected_clients.add(client_queue)

    def remove_client(self, client_queue):
        self.connected_clients.discard(client_queue)

    def create_endpoint(self):
        async def endpoint(request: Request):
            q = asyncio.Queue()
            self.add_client(q)
            return StreamingResponse(self.event_generator(request, q), media_type="text/event-stream")
        return endpoint

    async def producer(self):
        i = 1
        while True:
            await asyncio.sleep(10)
            print(f"📦 当前线程：{threading.current_thread().name}, 消息 {i}")
            await self.push_message(f"消息 {i}")
            i += 1





class SSESessionService:
    def __init__(self):
        self.global_queue = asyncio.Queue()
        self.client_groups: Dict[str, Set[asyncio.Queue]] = defaultdict(set)
        self.lock = asyncio.Lock()

    def add_client(self, client_queue: asyncio.Queue, group: str):
        self.client_groups[group].add(client_queue)

    def remove_client(self, client_queue: asyncio.Queue, group: str):
        self.client_groups[group].discard(client_queue)

    async def event_generator(self, request, client_queue: asyncio.Queue, group: str):
        yield ": keep-alive\n\n"
        try:
            while True:
                if await request.is_disconnected():
                    break
                try:
                    msg = await asyncio.wait_for(client_queue.get(), timeout=10)
                    yield f"data: {msg}\n\n"
                except asyncio.TimeoutError:
                    # yield ": keep-alive\n\n"
                    yield f"data: {json.dumps({'type': 'ping'})}\n\n"
        except asyncio.CancelledError:
            pass
        finally:
            async with self.lock:
                self.remove_client(client_queue, group)

    async def push_message(self, msg: dict):
        # 消息结构：{"group": "typeA", "data": "hello"}
        await self.global_queue.put(msg)

    async def push_message_default(self, msg: dict):
        # 消息结构：{"group": "typeA", "data": "hello"}
        await self.global_queue.put({"group": "default", "data": msg})

    async def broadcast_loop(self):
        while True:
            msg = await self.global_queue.get()
            group = msg.get("group")
            data = msg.get("data")
            if not group or data is None:
                continue  # 忽略非法消息
            async with self.lock:
                clients = self.client_groups.get(group, set()).copy()
            print(f"📣 广播到组 {group}: {data}, 客户端数量: {len(clients)}")
            for q in clients:
                await q.put(data)
    def create_endpoint(self):
        async def endpoint(request: Request, group: str = "default"):
            q = asyncio.Queue()
            self.add_client(q, group)
            return StreamingResponse(self.event_generator(request, q, group), media_type="text/event-stream")
        return endpoint

    async def producer(self):
        i = 0
        while True:
            await asyncio.sleep(10)
            i += 1
            msgA = {"group": "default", "data": f"A消息 {i}"}
            # msgB = {"group": "typeB", "data": f"B消息 {i}"}
            await self.push_message(msgA)
            # await self.push_message(msgB)


# @lru_cache()
# def get_sse_service():
#     sse_service = SSESessionService()
#     return sse_service

# from collections import defaultdict
# from typing import Dict, Set
# import asyncio
# import threading

# class SSEService:
#     def __init__(self):
#         # 改为多个 group 队列，每个 group 的 queue 是多个 client queue
#         self.client_groups: Dict[str, Set[asyncio.Queue]] = defaultdict(set)
#         self.lock = asyncio.Lock()

#     async def event_generator(self, request, client_queue, group: str):
#         try:
#             while True:
#                 if await request.is_disconnected():
#                     print(f"请求关闭! 组: {group}")
#                     break
#                 try:
#                     msg = await asyncio.wait_for(client_queue.get(), timeout=10)
#                     yield f"data: {msg}\n\n"
#                 except asyncio.TimeoutError:
#                     yield ": keep-alive\n\n"
#         except asyncio.CancelledError:
#             print("连接被取消")
#         finally:
#             print(f"finally 请求关闭! 组: {group}")
#             async with self.lock:
#                 self.client_groups[group].discard(client_queue)

#     async def push_message(self, msg: str, group: str):
#         async with self.lock:
#             queues = self.client_groups.get(group, set()).copy()
#         print(f"📢 向组 [{group}] 发送消息 {msg}，客户端数量: {len(queues)}")
#         for q in queues:
#             await q.put(msg)

#     def add_client(self, client_queue: asyncio.Queue, group: str):
#         self.client_groups[group].add(client_queue)

#     def remove_client(self, client_queue: asyncio.Queue, group: str):
#         self.client_groups[group].discard(client_queue)


# from fastapi import Request
# from fastapi.responses import StreamingResponse
# from fastapi import APIRouter

# sseController = APIRouter()

# @sseController.get("/sse")
# async def sse(request: Request, group: str = "default"):
#     q = asyncio.Queue()
#     sse_service = request.app.state.sse_service
#     sse_service.add_client(q, group)
#     return StreamingResponse(sse_service.event_generator(request, q, group), media_type="text/event-stream")



import shutil
import os

def delete_all_in_dir(path):
    print(f"delete file: {path}")
    if not os.path.exists(path):
        print(f"{path} 不存在")
        return
    
    for filename in os.listdir(path):
        file_path = os.path.join(path, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.remove(file_path)  # 删除文件或符号链接
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)  # 删除文件夹及其内容
        except Exception as e:
            print(f"删除 {file_path} 失败: {e}")



import asyncio
from fastapi import FastAPI, Request,Response, WebSocket
import os
from fastapi.responses import JSONResponse, StreamingResponse

from fastapi.staticfiles import StaticFiles
from starlette.websockets import WebSocketDisconnect
from brave.api.routes.project import project_api
from brave.api.routes.sample_result import sample_result
from brave.api.routes.file_parse_plot import file_parse_plot
from brave.api.routes.sample import sample
from brave.api.routes.analysis import analysis_api
from brave.api.routes.bio_database import bio_database
from brave.api.routes.pipeline import pipeline
from brave.api.routes.literature import literature_api
from brave.api.routes.sse import sseController
from brave.api.routes.namespace import namespace
from brave.api.routes.index import index_api

from brave.api.routes.file_operation import file_operation
from brave.api.routes.setting import setting_controller
from brave.api.config.config import get_settings
from fastapi.responses import FileResponse
from brave.app_manager import AppManager    
from brave.api.routes.container import container_controller
from brave.microbe.routes.entity_relation import entity_relation_api
from brave.microbe.routes.entity import entity_api
from brave.microbe.routes.study import  study_api
from brave.microbe.nlp.nlp import  nlp_api
from brave.api.routes.kegg import kegg_api
from brave.api.routes.component_store import component_store_api


import httpx
import websockets

def setup_routes(app: FastAPI,manager:AppManager):
  

    frontend_path = os.path.join(os.path.dirname(__file__), "frontend")
    app.mount("/assets", StaticFiles(directory=os.path.join(frontend_path, "build","assets")), name="assets")
    # frontend_path = os.path.join(os.path.dirname(__file__), "frontend")
    app.mount("/brave-api/img", StaticFiles(directory=os.path.join(frontend_path, "img")), name="img")
    settings = get_settings()
    app.mount("/brave-api/dir", StaticFiles(directory=settings.BASE_DIR, follow_symlink=True), name="base_dir")
    app.mount("/brave-api/work-dir", StaticFiles(directory=settings.WORK_DIR), name="work_dir")

    app.mount("/brave-api/literature/dir", StaticFiles(directory=os.path.join(settings.LITERATURE_DIR)), name="literature_dir")
    app.mount("/brave-api/pipeline-dir", StaticFiles(directory=os.path.join(settings.PIPELINE_DIR)), name="pipeline_dir")
    app.mount("/brave-api/store-dir", StaticFiles(directory=os.path.join(settings.STORE_DIR)), name="store_dir")

    app.include_router(sample_result,prefix="/brave-api")
    app.include_router(file_parse_plot,prefix="/brave-api")
    app.include_router(sample,prefix="/brave-api")
    app.include_router(analysis_api,prefix="/brave-api")
    app.include_router(bio_database,prefix="/brave-api")
    app.include_router(pipeline,prefix="/brave-api")
    app.include_router(literature_api,prefix="/brave-api")
    app.include_router(sseController,prefix="/brave-api")
    app.include_router(namespace,prefix="/brave-api")
    app.include_router(file_operation,prefix="/brave-api")
    app.include_router(setting_controller,prefix="/brave-api")
    app.include_router(container_controller,prefix="/brave-api")
    app.include_router(project_api,prefix="/brave-api")
    app.include_router(entity_api,prefix="/brave-api")
    app.include_router(entity_relation_api,prefix="/brave-api")
    app.include_router(study_api, prefix="/brave-api")
    app.include_router(nlp_api, prefix="/brave-api")
    app.include_router(kegg_api, prefix="/brave-api")
    app.include_router(index_api,prefix="/brave-api")
    app.include_router(component_store_api,prefix="/brave-api")



    app.get("/brave-api/sse-group")(manager.sse_service.create_endpoint())  
    endpoint = manager.ingress_manager.create_endpoint()
    if endpoint:
        app.post("/brave-api/ingress")(endpoint)
    # curl -X POST http://localhost:5005/brave-api/ingress -d '{"ingress_event": "workflow_log", "workflow_event":"flow_begin","workflow_id": "123", "message": "test"}'
    # app.state.sse_service = sse_service
    
    JUPYTER_SERVER_URL = "http://192.168.3.63:8888" # Replace with your Jupyter server URL

    # @app.api_route("/jupyter/{path:path}",  methods=["GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS"])
    # async def proxy_jupyter(request: Request, path: str):
    #     jupyter_url = f"{JUPYTER_SERVER_URL}/jupyter/{path}"

    #     async with httpx.AsyncClient() as client:
    #         proxy_req = client.build_request(
    #             request.method,
    #             jupyter_url,
    #             headers=request.headers.raw,
    #             content=await request.body()
    #         )
    #         response = await client.send(proxy_req, stream=True)
            
    #         return StreamingResponse(response.aiter_raw(), status_code=response.status_code, headers=dict(response.headers))
    
    JUPYTER_SERVER_HTTP = "http://192.168.3.63:8888"  # 你的 Jupyter HTTP 地址
    JUPYTER_SERVER_WS = "ws://192.168.3.63:8888"     # 你的 Jupyter WebSocket  
    # @app.api_route("/jupyter/{path:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS"])
    # async def proxy_jupyter(path: str, request: Request):
    #     async with httpx.AsyncClient() as client:
    #         # Construct the URL for the Jupyter server
    #         jupyter_url = f"{JUPYTER_SERVER_URL}/jupyter/{path}"

    #         # Forward the request
    #         response = await client.request(
    #             method=request.method,
    #             url=jupyter_url,
    #             headers=request.headers,
    #             content=await request.body(),
    #             params=request.query_params
    #         )

    #         # Return the response from Jupyter
    #         return Response(content=response.content, status_code=response.status_code, headers=response.headers)


    # @app.api_route("/jupyter/{full_path:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"])
    # async def proxy_http(full_path: str, request: Request):
    #     url = f"{JUPYTER_SERVER_HTTP}/jupyter/{full_path}"

    #     async with httpx.AsyncClient() as client:
    #         # 转发请求头，排除host等敏感头
    #         headers = dict(request.headers)
    #         headers.pop("host", None)

    #         response = await client.request(
    #             method=request.method,
    #             url=url,
    #             headers=headers,
    #             content=await request.body(),
    #             params=request.query_params
    #         )

    #     # 返回响应，过滤部分header避免冲突
    #     excluded_headers = ["content-encoding", "transfer-encoding", "connection"]
    #     headers = [(k, v) for k, v in response.headers.items() if k.lower() not in excluded_headers]
    #     return Response(content=response.content, status_code=response.status_code, headers=dict(headers))


    # @app.websocket("/jupyter/{full_path:path}")
    # async def proxy_ws(full_path: str, websocket: WebSocket):
    #     await websocket.accept()
    #     target_url = f"{JUPYTER_SERVER_WS}/jupyter/{full_path}"
    #     try:
    #         async with websockets.connect(target_url) as jupyter_ws:

    #             async def forward_ws_to_client():
    #                 try:
    #                     async for message in jupyter_ws:
    #                         await websocket.send_text(message)
    #                 except websockets.ConnectionClosed:
    #                     pass

    #             async def forward_client_to_ws():
    #                 try:
    #                     while True:
    #                         data = await websocket.receive_text()
    #                         await jupyter_ws.send(data)
    #                 except WebSocketDisconnect:
    #                     pass

    #             # 并发转发，等待任何一方断开
    #             await asyncio.wait([
    #                 forward_ws_to_client(),
    #                 forward_client_to_ws(),
    #             ], return_when=asyncio.FIRST_COMPLETED)

    #     except Exception as e:
    #         await websocket.close()
    
    # 启动后台广播任务
    # @app.on_event("startup")
   

        # asyncio.create_task(broadcast_loop())
        # await startup_process_event()
        # asyncio.create_task(producer())

    # @app.on_event("shutdown")
    # async def on_shutdown():
    #     print("✅ 关闭后台任务")
    #     for task in [producer_task, broadcast_task]:
    #         if task:
    #             task.cancel()


    # async def value_error_handler(request: Request, exc: Exception):
    #     return JSONResponse(
    #         status_code=400,
    #         content={"detail": str(exc), "type": "ValueError"}
    #     )

    # # 注册异常处理器
    # app.add_exception_handler(Exception, value_error_handler)
    # @app.exception_handler(Exception)
    @app.get("/favicon.ico")
    async def serve_favicon():
        favicon = os.path.join(frontend_path, "build/favicon.ico")
        return FileResponse(favicon)

    @app.get("/html/index.html")
    async def html():
        index_path = os.path.join(frontend_path, "html/index.html")
        return FileResponse(index_path)
    @app.get("/psycmicrograph.html")
    async def html():
        index_path = os.path.join(frontend_path, "build/psycmicrograph.html")
        return FileResponse(index_path)

    @app.get("/{full_path:path}")
    async def serve_frontend(full_path: str):
        index_path = os.path.join(frontend_path, "build/index.html")
        return FileResponse(index_path)




import uuid
from fastapi import APIRouter
import pandas as pd
import json

from sqlalchemy import select

from brave.api.config.db import get_engine
from brave.api.models.core import t_bio_database
from brave.api.schemas.bio_database import AddBioDatabase, QueryBiodatabase, UpdateBioDatabase
import brave.api.service.bio_database_service as bio_database_service

bio_database = APIRouter()

@bio_database.get("/fast-api/get_metaphlan_clade")
async def get_metaphlan_clade():
    with open("/ssd1/wy/workspace2/nextflow-fastapi/databases/clade.json") as f:
        data = json.load(f)
    return data


@bio_database.post("/list-bio-database",tags=["bio_database"])
async def list_bio_database(query: QueryBiodatabase):
    with get_engine().begin() as conn:
        result = bio_database_service.list_bio_database(conn,query)
    return result

@bio_database.get("/fast-api/get_bio_database_by_id")
async def get_bio_database_by_id(database_id: str):
    with get_engine().begin() as conn:
        result = bio_database_service.get_bio_database_by_id(conn,database_id)
    return result



@bio_database.post("/add-bio-database",tags=["bio_database"])
async def add_bio_database(database: AddBioDatabase):
    unique_id = str(uuid.uuid4())
    with get_engine().begin() as conn:
        conn.execute(t_bio_database.insert().values(database_id=unique_id,**database.model_dump()))
    return {"message": "Database added successfully"}


@bio_database.post("/update-bio-database",tags=["bio_database"])
async def update_bio_database(database: UpdateBioDatabase):
    with get_engine().begin() as conn:
        conn.execute(t_bio_database.update().where(t_bio_database.c.database_id == database.database_id).values(**database.model_dump()))
    return {"message": "Database updated successfully"}


@bio_database.delete("/delete-bio-database/{database_id}",tags=["bio_database"])
async def delete_bio_database(database_id: int):
    with get_engine().begin() as conn:
        conn.execute(t_bio_database.delete().where(t_bio_database.c.id == database_id))
    return {"message": "Database deleted successfully"}


# async def get_database_content_by_name(name: str):
#     with get_engine().begin() as conn:
#         stmt = select(t_bio_database).where(t_bio_database.c.name == name)
#         result = conn.execute(stmt).first()
        
#     return result


from typing import Optional
from pydantic import BaseModel

class AnalysisResultQuery(BaseModel):
    # id: Optional[int]
    analysis_method:Optional[list]=None
    component_ids:Optional[list]=None
    component_id:Optional[str]=None
    project:Optional[str]=None
    projectList:Optional[list]=None
    querySample:Optional[bool]=True
    # queryAnalysis:Optional[bool]=True
    analysis_type:Optional[str]=None
    ids:Optional[list]=None

class AnalysisResultParseModal(BaseModel):
    sample_id: Optional[str]
    project: Optional[str]
    component_id: Optional[str]
    analysis_result_hash: Optional[str]
    file_name: Optional[str]

class AnalysisResult(BaseModel):
    id: Optional[int]
    sample_name: Optional[str]
    # sample_key: Optional[str]
    sample_id: Optional[str]
    analysis_name: Optional[str]=None
    analysis_result_id: Optional[str]
    analysis_key: Optional[str]
    analysis_method: Optional[str]
    software: Optional[str]
    content: Optional[str]
    file_name: Optional[str]
    # analysis_version: Optional[str]
    content_type: Optional[str]
    project: Optional[str]
    request_param: Optional[str]
    analysis_id: Optional[str]
    # sample_group: Optional[str]
    # sample_group_name: Optional[str]
    analysis_type: Optional[str]
    create_date: Optional[str]
    # sample_source: Optional[str]
    # host_disease: Optional[str]
    component_id: Optional[str]
    component_name: Optional[str]
    component_label: Optional[str]
    analysis_result_hash: Optional[str]
    

class ParseImportData(BaseModel):
    content: str
    
    # suffix: str
    
# stmt =  select(analysis_result, samples).select_from(analysis_result.outerjoin(samples,samples.c.sample_key==analysis_result.c.analysis_key))

class ImportData(BaseModel):
    component_id:str
    project: str
    content: str
    sample_name: str
    sample_source: str
    file_name: Optional[str]=None
    # content_type:Optional[str]="json"
    # analysis_type:Optional[str]="import_data"

class UpdateAnalysisResult(BaseModel):
    id: int
    analysis_key: str


class BindSample(BaseModel):
    analysis_result_id: str
    sample_id: str


# brave/workflow_events/__init__.py
# 空文件，用于模块识别


# brave/workflow_events/types.py
from typing import TypedDict, Optional

class EventMessage(TypedDict):
    workflow_id: str
    event: str
    timestamp: str
    msg: Optional[str]
    level: Optional[str]


# brave/workflow_events/pubsub.py
from collections import defaultdict
from typing import Callable, Awaitable, Dict, List, Any

Subscriber = Callable[[Any], Awaitable[None]]

class PubSubManager:
    def __init__(self):
        self.subscribers: Dict[str, List[Subscriber]] = defaultdict(list)

    def subscribe(self, topic: str, callback: Subscriber):
        self.subscribers[topic].append(callback)

    async def publish(self, topic: str, message: Any):
        for cb in self.subscribers.get(topic, []):
            await cb(message)


# brave/workflow_events/workflow_queue.py
import asyncio
from collections import defaultdict
from typing import Dict, Any

class WorkflowQueueManager:
    def __init__(self, pubsub):
        self.queues: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self.pubsub = pubsub
        self._started = False

    async def put(self, workflow_id: str, msg: dict):
        await self.queues[workflow_id].put(msg)

    async def _consume_loop(self, workflow_id: str):
        queue = self.queues[workflow_id]
        while True:
            msg = await queue.get()
            await self.pubsub.publish(msg.get("event", "default"), msg)

    def start(self):
        if self._started:
            return
        self._started = True
        for workflow_id in self.queues:
            asyncio.create_task(self._consume_loop(workflow_id))


# brave/workflow_events/uds_listener.py
import os
import asyncio
import socket
import json
from .event_router import EventRouter

class UDSListener:
    def __init__(self, uds_path: str, router: EventRouter):
        self.uds_path = uds_path
        self.router = router

    async def start(self):
        if os.path.exists(self.uds_path):
            os.remove(self.uds_path)

        server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        server.bind(self.uds_path)
        server.listen(5)
        server.setblocking(False)

        loop = asyncio.get_running_loop()
        print(f"[UDS] Listening at {self.uds_path}")

        while True:
            client, _ = await loop.sock_accept(server)
            asyncio.create_task(self._handle_client(client))

    async def _handle_client(self, client: socket.socket):
        loop = asyncio.get_running_loop()
        with client:
            while True:
                try:
                    data = await loop.sock_recv(client, 4096)
                    if not data:
                        break
                    try:
                        msg = json.loads(data.decode())
                        await self.router.dispatch(msg)
                    except Exception as e:
                        print(f"[UDS] Message error: {e}")
                except Exception as e:
                    print(f"[UDS] Client error: {e}")
                    break


# brave/workflow_events/event_router.py
from .workflow_queue import WorkflowQueueManager
from .pubsub import PubSubManager

class EventRouter:
    def __init__(self, pubsub: PubSubManager, wq_manager: WorkflowQueueManager):
        self.pubsub = pubsub
        self.wq_manager = wq_manager

    async def dispatch(self, msg: dict):
        workflow_id = msg.get("workflow_id", "global")
        await self.wq_manager.put(workflow_id, msg)
        self.wq_manager.start()


# brave/workflow_events/sse.py
from fastapi import Request
from fastapi.responses import StreamingResponse
import asyncio

class SSEManager:
    def __init__(self, wq_manager):
        self.wq_manager = wq_manager

    def create_endpoint(self):
        async def endpoint(workflow_id: str, request: Request):
            queue = self.wq_manager.queues[workflow_id]

            async def event_generator():
                while True:
                    if await request.is_disconnected():
                        break
                    try:
                        msg = await asyncio.wait_for(queue.get(), timeout=10)
                        yield f"data: {msg}\n\n"
                    except asyncio.TimeoutError:
                        yield ": keep-alive\n\n"

            return StreamingResponse(event_generator(), media_type="text/event-stream")

        return endpoint


# brave/workflow_events/manager.py
from .pubsub import PubSubManager
from .workflow_queue import WorkflowQueueManager
from .uds_listener import UDSListener
from .event_router import EventRouter
from .sse import SSEManager

class WorkflowEventSystem:
    def __init__(self, uds_path="/tmp/brave.sock"):
        self.pubsub = PubSubManager()
        self.queue_manager = WorkflowQueueManager(self.pubsub)
        self.router = EventRouter(self.pubsub, self.queue_manager)
        self.uds_listener = UDSListener(uds_path, self.router)
        self.sse = SSEManager(self.queue_manager)

    async def start(self):
        await self.uds_listener.start()

    def sse_endpoint(self):
        return self.sse.create_endpoint()

    def register_subscriber(self, topic: str, callback):
        self.pubsub.subscribe(topic, callback)



import asyncio
from typing import Dict, List, Callable, Coroutine, Optional
from fastapi import FastAPI, Request, Depends
from starlette.responses import StreamingResponse

app = FastAPI()
# 观察者模式（Observer Pattern）
# 你将每个日志源（Pod/Container）视为主题（Subject）

# 每个客户端连接（SSE）是一个观察者（Observer）

# 新用户只需注册订阅已有流，不再创建重复连接

# 优点：支持多个用户复用同一日志流


# --- LogStream 类 ---
class LogStream:
    def __init__(self, source_id: str, log_source: Callable[[], Coroutine]):
        self.source_id = source_id
        self.log_source_func = log_source
        self.subscribers: List[asyncio.Queue] = []
        self.task: Optional[asyncio.Task] = None
        self.running = False

    def subscribe(self) -> asyncio.Queue:
        q = asyncio.Queue()
        self.subscribers.append(q)
        return q

    def unsubscribe(self, q: asyncio.Queue):
        if q in self.subscribers:
            self.subscribers.remove(q)

    async def start(self):
        if self.running:
            return
        self.running = True

        async def stream_logs():
            try:
                async for line in self.log_source_func():
                    if not self.subscribers:
                        break
                    for q in list(self.subscribers):
                        await q.put(line)
            except Exception as e:
                print(f"Log stream error: {e}")
            finally:
                self.running = False

        self.task = asyncio.create_task(stream_logs())

    async def close(self):
        self.subscribers.clear()
        if self.task:
            self.task.cancel()
            self.task = None
        self.running = False


# --- 单例注册中心 ---
class LogStreamRegistry:
    def __init__(self):
        self.streams: Dict[str, LogStream] = {}

    def get_or_create(self, source_id: str, log_source: Callable[[], Coroutine]) -> LogStream:
        if source_id not in self.streams:
            self.streams[source_id] = LogStream(source_id, log_source)
        return self.streams[source_id]

    def remove(self, source_id: str):
        self.streams.pop(source_id, None)

log_registry = LogStreamRegistry()


# --- 模拟日志源（可替换为 docker/k8s 实现） ---
async def fake_log_generator():
    for i in range(100):
        await asyncio.sleep(0.5)
        yield f"log line {i}"


# --- SSE 接口 ---
@app.get("/logs/{source_id}")
async def stream_logs(request: Request, source_id: str):
    stream = log_registry.get_or_create(source_id, fake_log_generator)
    q = stream.subscribe()
    await stream.start()

    async def event_gen():
        try:
            while True:
                if await request.is_disconnected():
                    break
                try:
                    line = await asyncio.wait_for(q.get(), timeout=10)
                    yield f"data: {line}\n\n"
                except asyncio.TimeoutError:
                    yield ": keep-alive\n\n"
        except asyncio.CancelledError:
            pass
        finally:
            stream.unsubscribe(q)

    return StreamingResponse(event_gen(), media_type="text/event-stream")


# --- 后台清理任务 ---
async def cleanup_task():
    while True:
        to_remove = []
        for sid, stream in log_registry.streams.items():
            if not stream.subscribers:
                await stream.close()
                to_remove.append(sid)
        for sid in to_remove:
            log_registry.remove(sid)
        await asyncio.sleep(30)

@app.on_event("startup")
async def start_cleanup():
    asyncio.create_task(cleanup_task())



from watchfiles import awatch
# from brave.api.routes.sse import global_queue
from datetime import datetime
import asyncio
from importlib.resources import files, as_file
from importlib import import_module
from brave.api.config.db import get_engine
from brave.api.models.core import analysis
import psutil
import logging
from sqlalchemy import  select, update
import inspect

# logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

listener_files = files("brave.api.listener")
listener_files = [
    item.stem
    for item in listener_files.iterdir() 
    if item.is_file() and item.name.endswith(".py") and item.name != "__init__.py" and item.name.startswith("file")     
]

async def execute_listener(func, args):
    if isinstance(listener_files, list) and len(listener_files) > 0:
        for name in listener_files:
            full_module = f"brave.api.listener.{name}"
            mod = import_module(full_module)
            if hasattr(mod, func):
                run_func = getattr(mod, func)
                if inspect.iscoroutinefunction(run_func):
                    asyncio.create_task(run_func(**args))
                else:
                    await asyncio.to_thread(run_func, **args)


# 文件变更监控任务
async def watch_folder(path: str):
   
    async for changes in awatch(path, recursive=False, step=3000):
        for change, file_path in changes:
            await execute_listener("file_change", {"change":change, "file_path":file_path})
            # msg = f"{change.name.upper()} {file_path}"
            # if len(listener_files) >0:
            #     for name in listener_files:
            #         full_module = f"brave.api.listener.{name}"
            #         mod = import_module(full_module)
            #         if hasattr(mod, "run"):
            #             # await mod.run(change.name.upper(),file_path)
            #             run_func = mod.run
            #             if inspect.iscoroutinefunction(run_func):
            #                 asyncio.create_task(run_func(change.name.upper(), file_path))
            #             else:
            #                 await asyncio.to_thread(run_func, change.name.upper(), file_path)

            # await global_queue.put(msg)


queue_process = asyncio.Queue()
queue_lock = asyncio.Lock()  # 保证数据库更新和队列操作安全

async def startup_process_event():
    with get_engine().begin() as conn:
        stmt = select(analysis).where(analysis.c.process_id != None)
        results = conn.execute(stmt).all()
        for row in results:
            item = dict(row._mapping)
            await queue_process.put(item)
    logger.info(f"队列初始化完毕，任务数：{queue_process.qsize()}")   
    asyncio.create_task(check_process_worker())

CHECK_INTERVAL = 5  # 秒，检查频率
async def check_process_worker():
    while True:
        item = await queue_process.get()
        process_id = item.get("process_id")
        analysis_id = item.get("analysis_id")
        logger.info(f"检查分析任务 id={analysis_id}, pid={process_id}")
        try:
            pid_int = int(process_id)
            proc = psutil.Process(pid_int)
            proc_name = proc.name().lower()
            if "bash" not in proc_name:
                raise psutil.NoSuchProcess(f"进程 {process_id} 不是 nextflow")

        except (psutil.NoSuchProcess, ValueError) as e:
            logger.warning(f"进程 {process_id} 不存在或非 nextflow，清理数据库: {e}")
            # 更新数据库，将 process_id 设为 None
            async with queue_lock:
                with get_engine().begin() as conn:
                    stmt = (
                        update(analysis)
                        .where(analysis.c.analysis_id == analysis_id)
                        .values(process_id=None)
                    )
                    conn.execute(stmt)
                    conn.commit()
                logger.info(f"清理完成 analysis id={analysis_id}")
                await execute_listener("process_end", {"analysis_id":analysis_id})
                # 不重新入队，直接丢弃任务
        else:
            # 进程存在且符合，延迟后重新入队
            await asyncio.sleep(CHECK_INTERVAL)
            await queue_process.put(item)     
        finally:
            queue_process.task_done()





# async def debounce_handler(path, debounce_seconds=2):
#     last_event_time = None
#     debounce_task = None

#     async for changes in awatch(path, recursive=False, step=2000):
#         now = datetime.now()
#         print(f"[{now.strftime('%H:%M:%S')}] 检测到文件变化：{changes}")

#         if debounce_task and not debounce_task.done():
#             debounce_task.cancel()
#             print("取消上一次定时任务")

#         debounce_task = asyncio.create_task(wait_and_handle(debounce_seconds, changes))


# async def wait_and_handle(debounce_seconds, changes):
#     try:
        
#         print(f"✅ {debounce_seconds} 秒后触发处理：{changes}")
#         # 在这里处理变化，比如重新执行某个流程
#         for change, file_path in changes:
#             msg = f"{change.name.upper()} {file_path}"
#             await global_queue.put(msg)

#         await asyncio.sleep(debounce_seconds)

#     except asyncio.CancelledError:
#         # 被新的变化打断
#         pass


# models.py
from typing import List,Optional
from pydantic import BaseModel
import hashlib
import json
           
class JobSpec(BaseModel):
    job_id: str
    command: List[str]
    output_dir: str
    command_log_path: str
    # image: str = ""
    # env: dict = {}
    # resources: dict = {}

    # def fingerprint(self) -> str:
    #     key = {
    #         "cmd": self.command,
    #         "img": self.image,
    #         "env": self.env,
    #         "res": self.resources,
    #     }
    #     return hashlib.sha256(json.dumps(key, sort_keys=True).encode()).hexdigest()
class LocalJobSpec(JobSpec):
    process_id: Optional[int]=None


class DockerJobSpec(JobSpec):
    container_id: str
    resources: dict
    run_type:str
    change_uid:bool


class JobStatus(BaseModel):
    job_id: str
    state: str
    detail: str = ""



from fastapi import Request
from fastapi.responses import StreamingResponse
import asyncio
from brave.api.core.workflow_queue import WorkflowQueueManager
class WorkflowSSEManager:
    def __init__(self, workflow_queue_manager: WorkflowQueueManager):
        self.workflow_queue_manager = workflow_queue_manager
    def create_endpoint(self):
        async def endpoint(workflow_id: str, request: Request):
            queue = asyncio.Queue()

            # 订阅 pubsub，收到新消息后放入本地队列
            def subscriber(msg):
                queue.put_nowait(msg)

            self.workflow_queue_manager.register_subscriber(workflow_id, subscriber)

            async def event_generator():
                try:
                    while True:
                        if await request.is_disconnected():
                            break
                        try:
                            msg = await asyncio.wait_for(queue.get(), timeout=10)
                            yield f"data: {msg}\n\n"
                        except asyncio.TimeoutError:
                            yield ": keep-alive\n\n"
                finally:
                    # 取消订阅
                    self.workflow_queue_manager.pubsub.unsubscribe(workflow_id, subscriber)

            return StreamingResponse(event_generator(), media_type="text/event-stream")

        return endpoint
    # def create_endpoint(self):
    #     async def endpoint(workflow_id: str, request: Request):
    #         queue = self.workflow_queue_manager.get(workflow_id)

    #         async def event_generator():
    #             while True:
    #                 if await request.is_disconnected():
    #                     break
    #                 try:
    #                     msg = await asyncio.wait_for(queue.queue.get(), timeout=10)
    #                     yield f"data: {msg}\n\n"
    #                 except asyncio.TimeoutError:
    #                     yield ": keep-alive\n\n"

    #         return StreamingResponse(event_generator(), media_type="text/event-stream")

    #     return endpoint







from brave.api.config.db import get_engine
from brave.api.models.core import t_bio_database
from brave.api.schemas.bio_database import QueryBiodatabase
from sqlalchemy import select

def list_bio_database(conn,query: QueryBiodatabase):
    condition = []
    if query.database_id:
        condition.append(t_bio_database.c.database_id == query.database_id)
    elif query.type:
        condition.append(t_bio_database.c.type == query.type)
    elif query.name:
        condition.append(t_bio_database.c.name == query.name)
    elif query.path:
        condition.append(t_bio_database.c.path == query.path)
    elif query.type_list:
        condition.append(t_bio_database.c.type.in_(query.type_list))


    stmt = select(t_bio_database).where(*condition)
    result = conn.execute(stmt).mappings().all()
    return result


def get_bio_database_by_id(conn,database_id: str):
    stmt = select(t_bio_database).where(t_bio_database.c.database_id == database_id)
    result = conn.execute(stmt).mappings().first()
    return result



# brave/workflow_events/__init__.py
# 空文件，用于模块识别


# brave/workflow_events/types.py
from typing import TypedDict, Optional

class EventMessage(TypedDict):
    workflow_id: str
    event: str
    timestamp: str
    msg: Optional[str]
    level: Optional[str]


# brave/workflow_events/pubsub.py
from collections import defaultdict
from typing import Callable, Awaitable, Dict, List, Any

Subscriber = Callable[[Any], Awaitable[None]]

class PubSubManager:
    def __init__(self):
        self.subscribers: Dict[str, List[Subscriber]] = defaultdict(list)

    def subscribe(self, topic: str, callback: Subscriber):
        self.subscribers[topic].append(callback)

    async def publish(self, topic: str, message: Any):
        for cb in self.subscribers.get(topic, []):
            await cb(message)


# brave/workflow_events/workflow_queue.py
import asyncio
from collections import defaultdict
from typing import Dict, Any
import time
from dataclasses import dataclass, field

@dataclass
class WorkflowQueue:
    queue: asyncio.Queue
    task: asyncio.Task
    last_active: float = field(default_factory=time.time)
    subscribers: int = 0

class WorkflowQueueManager:
    def __init__(self, pubsub):
        self.workflow_map: Dict[str, WorkflowQueue] = {}
        self.pubsub = pubsub

    def register(self, workflow_id: str):
        if workflow_id not in self.workflow_map:
            queue = asyncio.Queue()
            task = asyncio.create_task(self._consume_loop(workflow_id, queue))
            self.workflow_map[workflow_id] = WorkflowQueue(queue=queue, task=task)

    async def put(self, workflow_id: str, msg: dict):
        self.register(workflow_id)
        wfq = self.workflow_map[workflow_id]
        wfq.last_active = time.time()
        await wfq.queue.put(msg)

    def get(self, workflow_id: str) -> WorkflowQueue:
        return self.workflow_map[workflow_id]

    async def _consume_loop(self, workflow_id: str, queue: asyncio.Queue):
        while True:
            msg = await queue.get()
            try:
                await self.pubsub.publish(workflow_id, msg)
            except Exception as e:
                print(f"[Consumer ERROR] workflow {workflow_id}: {e}")

    async def cleanup(self, timeout: int = 300):
        now = time.time()
        to_delete = []
        for wf_id, wfq in self.workflow_map.items():
            if wfq.subscribers == 0 and wfq.queue.empty() and (now - wfq.last_active > timeout):
                wfq.task.cancel()
                to_delete.append(wf_id)
        for wf_id in to_delete:
            del self.workflow_map[wf_id]
            print(f"[Cleanup] Removed workflow {wf_id}")


# brave/workflow_events/uds_listener.py
import os
import asyncio
import json
from .event_router import EventRouter

class UDSListener:
    def __init__(self, uds_path: str, router: EventRouter):
        self.uds_path = uds_path
        self.router = router

    async def start(self):
        if os.path.exists(self.uds_path):
            os.remove(self.uds_path)

        server = await asyncio.start_unix_server(self.handle_client, path=self.uds_path)
        print(f"[UDS] Listening at {self.uds_path}")
        async with server:
            await server.serve_forever()

    async def handle_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        try:
            while line := await reader.readline():
                try:
                    msg = json.loads(line.decode().strip())
                    await self.router.dispatch(msg)
                except Exception as e:
                    print(f"[UDS] Message error: {e}")
        except Exception as e:
            print(f"[UDS] Client error: {e}")
        finally:
            writer.close()
            await writer.wait_closed()


# brave/workflow_events/event_router.py
from .workflow_queue import WorkflowQueueManager
from .pubsub import PubSubManager

class EventRouter:
    def __init__(self, pubsub: PubSubManager, wq_manager: WorkflowQueueManager):
        self.pubsub = pubsub
        self.wq_manager = wq_manager

    async def dispatch(self, msg: dict):
        workflow_id = msg.get("workflow_id", "global")
        await self.wq_manager.put(workflow_id, msg)


# brave/workflow_events/sse.py
from fastapi import Request
from fastapi.responses import StreamingResponse
import asyncio
import json
import time

class SSEManager:
    def __init__(self, wq_manager):
        self.wq_manager = wq_manager

    def create_endpoint(self):
        async def sse_endpoint(workflow_id: str, request: Request):
            self.wq_manager.register(workflow_id)
            wfq = self.wq_manager.get(workflow_id)
            wfq.subscribers += 1
            wfq.last_active = time.time()

            async def event_stream():
                try:
                    while True:
                        if await request.is_disconnected():
                            break
                        msg = await wfq.queue.get()
                        yield f"data: {json.dumps(msg)}\n\n"
                finally:
                    wfq.subscribers -= 1
                    wfq.last_active = time.time()

            return StreamingResponse(event_stream(), media_type="text/event-stream")

        return sse_endpoint


# brave/workflow_events/manager.py
from .pubsub import PubSubManager
from .workflow_queue import WorkflowQueueManager
from .event_router import EventRouter
from .uds_listener import UDSListener
from .sse import SSEManager

class WorkflowEventSystem:
    def __init__(self, uds_path="/tmp/brave.sock"):
        self.pubsub = PubSubManager()
        self.queue_manager = WorkflowQueueManager(self.pubsub)
        self.router = EventRouter(self.pubsub, self.queue_manager)
        self.uds_listener = UDSListener(uds_path, self.router)
        self.sse = SSEManager(self.queue_manager)

    async def start(self):
        await self.uds_listener.start()

    def sse_endpoint(self):
        return self.sse.create_endpoint()

    def register_subscriber(self, topic: str, callback):
        self.pubsub.subscribe(topic, callback)



from typing import Optional
from pydantic import BaseModel

class QueryBiodatabase(BaseModel):
    database_id:Optional[str] = None
    name:Optional[str] = None
    path:Optional[str] = None
    type:Optional[str] = None
    type_list:Optional[list[str]] = None

class AddBioDatabase(BaseModel):
    name:str
    path:str
    db_index:Optional[str] = None
    type:str

class UpdateBioDatabase(BaseModel):
    database_id:str
    name:Optional[str] = None
    path:Optional[str] = None
    type:Optional[str] = None
    db_index:Optional[str] = None


from functools import reduce
from random import sample
import uuid
from fastapi import APIRouter,Depends
from sqlalchemy.orm import Session
import threading

# from brave.api.config.db import conn
# from models.user import users
from typing import List
from starlette.status import HTTP_204_NO_CONTENT
from sqlalchemy import func, select
from brave.api.models.orm import SampleAnalysisResult
import glob
import importlib
import os
import json
from brave.api.config.db import get_db_session
from sqlalchemy import and_,or_
from brave.api.schemas.analysis_result import AnalysisResultQuery,AnalysisResult, ImportData, ParseImportData,UpdateAnalysisResult,BindSample
from brave.api.models.core import samples,analysis_result
from brave.api.config.db import get_engine
import inspect
from fastapi import HTTPException
from brave.api.service.pipeline  import get_default_module, get_all_module,get_pipeline_dir
import threading
import brave.api.service.analysis_result_service as analysis_result_service
import brave.api.service.pipeline as pipeline_service
import re
from brave.api.utils.from_glob_get_file import from_glob_get_file
from brave.api.schemas.sample import AddSampleMetadata,UpdateSampleMetadata
import brave.api.service.sample_service as sample_service
from brave.api.service.analysis_result_parse import AnalysisResultParse
from fastapi import Depends
from dependency_injector.wiring import inject, Provide
from brave.app_container import AppContainer
from collections import defaultdict
sample_result = APIRouter()
# key = Fernet.generate_key()
# f = Fernet(key)


# def get_all_subclasses(cls):
#     subclasses = set(cls.__subclasses__())
#     for subclass in subclasses.copy():
#         subclasses.update(get_all_subclasses(subclass))
#     return subclasses




def update_or_save_result(analysis_key,sample_name, software, content_type, content, db, project, verison, analysis_method,analysis_name,analysis_id):
        sampleAnalysisResult = db.query(SampleAnalysisResult) \
        .filter(and_(SampleAnalysisResult.analysis_method == analysis_method,\
                SampleAnalysisResult.analysis_version == verison, \
                SampleAnalysisResult.analysis_key == analysis_key, \
                SampleAnalysisResult.project == project \
            )).first()
        if sampleAnalysisResult:
            sampleAnalysisResult.sample_name = sample_name
            sampleAnalysisResult.content = content
            sampleAnalysisResult.sample_key=sample_name
            sampleAnalysisResult.content_type = content_type
            sampleAnalysisResult.analysis_name = analysis_name
            sampleAnalysisResult.analysis_id = analysis_id
            sampleAnalysisResult.analysis_type="upstream"
            # sampleAnalysisResult.log_path = log_path
            sampleAnalysisResult.software = software
            db.commit()
            db.refresh(sampleAnalysisResult)
            print(">>>>更新: ",sample_name, software, content_type)
        else:
            sampleAnalysisResult = SampleAnalysisResult(analysis_method=analysis_method, \
                analysis_version=verison, \
                sample_name=sample_name, \
                content_type=content_type, \
                analysis_name=analysis_name, \
                analysis_key=analysis_key, \
                analysis_id=analysis_id, \
                analysis_type="upstream", \
                # log_path=log_path, \
                software=software, \
                project=project, \
                sample_key=sample_name, \
                content=content \
                    )
            db.add(sampleAnalysisResult)
            db.commit()
            print(">>>>新增: ",sample_name, software, content_type)



def parse_result_one(analysis_method,module,dir_path,project,verison,analysis_id=-1):
    parse = getattr(module, "parse")
    res = parse(dir_path)
    if hasattr(module,"get_analysis_method"):
        get_analysis_method = getattr(module, "get_analysis_method")
        analysis_method = get_analysis_method()
        print(f">>>>>更改分析名称: {analysis_method}")
    analysis_name = analysis_method
    if hasattr(module,"get_analysis_name"):
        get_analysis_name = getattr(module, "get_analysis_name")
        analysis_name = get_analysis_name()
        
    with get_db_session() as db:
        if len(res) >0:
            # print(res[0])
            if len(res[0]) == 4:
                for analysis_key,software,content_type,content in res:
                    update_or_save_result(analysis_key,analysis_key, software, content_type, content, db, project, verison, analysis_method,analysis_name,analysis_id)
            elif len(res[0]) == 5:
                for analysis_key,sample_name,software,content_type,content in res:
                    update_or_save_result(analysis_key,sample_name, software, content_type, content, db, project, verison, analysis_method,analysis_name,analysis_id)
            # print(sample_name)

def parse_result(dir_path,project,verison):
    # pipeline_dir = get_pipeline_dir()
    # py_files = [f for f in os.listdir(f"{pipeline_dir}/*/py_sample_result_parse/*") if f.endswith('.py')]
    # py_files = glob.glob(f"{pipeline_dir}/*/py_sample_result_parse/*.py")
    py_files ={} # get_all_module("py_sample_result_parse")
    for key,py_module in py_files.items():
        # module_name = py_file[:-3]  # 去掉 `.py` 后缀，获取模块名

       
        # if module_name not in all_module:
        #     raise HTTPException(status_code=500, detail=f"py_sample_result_parse: {module_name}没有找到!")
        # py_module = all_module[module_name]
        module = importlib.import_module(py_module)
        support_analysis_method = getattr(module, "support_analysis_method")
        analysis_method = support_analysis_method()

        
        if dir_path.endswith(analysis_method):
            print(f">>>>>找到分析名称 {analysis_method} 的分析结果")
            parse_result_one(analysis_method,module,dir_path,project,verison)
            # parse = getattr(module, "parse")
            # res = parse(dir_path)
            # if hasattr(module,"get_analysis_method"):
            #     get_analysis_method = getattr(module, "get_analysis_method")
            #     analysis_method = get_analysis_method()
            #     print(f">>>>>更改分析名称: {analysis_method}")
            # analysis_name = analysis_method
            # if hasattr(module,"get_analysis_name"):
            #     get_analysis_name = getattr(module, "get_analysis_name")
            #     analysis_name = get_analysis_name()
             
            # with get_db_session() as db:
            #     if len(res) >0:
            #         # print(res[0])
            #         if len(res[0]) == 4:
            #             for analysis_key,software,content_type,content in res:
            #                 update_or_save_result(analysis_key,analysis_key, software, content_type, content, db, project, verison, analysis_method,analysis_name)
            #         elif len(res[0]) == 5:
            #              for analysis_key,sample_name,software,content_type,content in res:
            #                 update_or_save_result(analysis_key,sample_name, software, content_type, content, db, project, verison, analysis_method,analysis_name)
            #         # print(sample_name)


@sample_result.get("/sample-parse-result-test-hexiaoyan",tags=['analsyis_result'])
async def parse_result_restful_test1():
    # base_path ="/ssd1/wy/workspace2/test/test_workspace/result/V1.0"
    # verison = "V1.0"
    # project="test"
    base_path ="/ssd1/wy/workspace2/hexiaoyan/hexiaoyan_workspace2/output"
    verison = "V1.0"
    project="hexiaoyan"
    
    dir_list = glob.glob(f"{base_path}/*",recursive=True)
    for dir_path in dir_list:
        parse_result(dir_path,project,verison)
    return {"msg":"success"}

@sample_result.get("/sample-parse-result-test",tags=['analsyis_result'])
async def parse_result_restful_test2():
    # base_path ="/ssd1/wy/workspace2/test/test_workspace/result/V1.0"
    # verison = "V1.0"
    # project="test"
    base_path ="/ssd1/wy/workspace2/leipu/leipu_workspace2/output"
    verison = "V1.0"
    project="leipu"
    
    dir_list = glob.glob(f"{base_path}/*",recursive=True)
    for dir_path in dir_list:
        parse_result(dir_path,project,verison)
    return {"msg":"success"}

@sample_result.get("/sample-parse-result-test-leipu-meta",tags=['analsyis_result'])
async def parse_result_restful_test3():
    # base_path ="/ssd1/wy/workspace2/test/test_workspace/result/V1.0"
    # verison = "V1.0"
    # project="test"
    base_path ="/ssd1/wy/workspace2/leipu/leipu_workspace_meta/output"
    verison = "V1.0"
    project="leipu"
    
    dir_list = glob.glob(f"{base_path}/*",recursive=True)
    for dir_path in dir_list:
        parse_result(dir_path,project,verison)
    return {"msg":"success"}

@sample_result.get("/sample-parse-result")
async def parse_result_restful(base_path,verison,project):
    # base_path ="/ssd1/wy/workspace2/test/test_workspace/result"
    # verison = "V1.0"
    
    dir_list = glob.glob(f"{base_path}/{verison}/*",recursive=True)
    for dir_path in dir_list:
        parse_result(dir_path,project,verison)
    return {"msg":"success"}

# def find_analyais_result_by_ids( value):
#     with get_engine().begin() as conn:
#         result_dict = analysis_result_service.find_analyais_result_by_ids(conn,value)
#     return result_dict

def find_analyais_result(analysisResultQuery:AnalysisResultQuery):
    with get_engine().begin() as conn:
        result_dict = analysis_result_service.find_analyais_result(conn,analysisResultQuery)
    return result_dict

def get_analysis_result_metadata(item):
    if item["metadata"]:
        metadata = json.loads(item["metadata"])
        prefix = ""
        if item["sample_source"]:
            prefix = f"{item['sample_source']}-"
        metadata = {k:f"{prefix}{v}" for k,v in metadata.items() if v is not None}
        item = {**metadata,**item}
        del item["metadata"]
    return item
@sample_result.post(
    "/analysis-result/list-analysis-result",
    # response_model=List[AnalysisResult]
    )
async def list_analysis_result(analysisResultQuery:AnalysisResultQuery):
    result_dict = find_analyais_result(analysisResultQuery)
    result_dict = [get_analysis_result_metadata(item) for item in result_dict]
            
        # if item["metadata_form"]:
        #     item["metadata_form"] = json.loads(item["metadata_form"])
    # grouped = defaultdict(list)
    # for item in result_dict:
    #     grouped[item["component_id"]].append(item)
    return result_dict

@sample_result.post(
    "/analysis-result/list-analysis-result-grouped",
    # response_model=List[AnalysisResult]
    )
async def list_analysis_result(analysisResultQuery:AnalysisResultQuery):
    result_dict = find_analyais_result(analysisResultQuery)
    result_dict = [get_analysis_result_metadata(item) for item in result_dict]
            
        # if item["metadata_form"]:
        #     item["metadata_form"] = json.loads(item["metadata_form"])
    grouped = defaultdict(list)
    for item in result_dict:
        item["label"] = item["sample_name"]
        item["value"] = item["id"]
        grouped[item["component_id"]].append(item)
    return grouped


@sample_result.post(
    "/fast-api/find-analyais-result-by-analysis-method",
    # response_model=List[AnalysisResult]
    )
async def find_analyais_result_by_analysis_method(analysisResultQuery:AnalysisResultQuery):
    return find_analyais_result(analysisResultQuery)
    # with get_db_session() as session:
    #     analysis_result =  session.query(SampleAnalysisResult,Sample) \
    #         .outerjoin(Sample, SampleAnalysisResult.sample_key == Sample.sample_key) \
    #         .filter(and_( \
    #             SampleAnalysisResult.analysis_method.in_(analysisResultQuery.analysis_method), \
    #             SampleAnalysisResult.project == analysisResultQuery.project \
    #         )) \
    #             .all()

    #     for item in analysis_result:
    #         if item.content_type=="json":
    #             item.content = json.loads(item.content)
            # print()
    # print(f"find_analyais_result_by_analysis_method 当前线程：{threading.current_thread().name}")
  
    # return {"aa":"aa"}


@sample_result.delete(
    "/analyais-result/delete-by-id/{analysis_result_id}",  
    status_code=HTTP_204_NO_CONTENT)
@inject
async def delete_analysis_result(
    analysis_result_id: str,
    analysis_result_parse_service:AnalysisResultParse = Depends(Provide[AppContainer.analysis_result_parse_service])):

    with get_engine().begin() as conn:
        find_analysis_result = analysis_result_service.find_by_analysis_result_id(conn,analysis_result_id)
        if not find_analysis_result:
            raise HTTPException(status_code=500, detail=f"分析结果{analysis_result_id}不存在!")
        analysis_id = find_analysis_result.analysis_id
        conn.execute(analysis_result.delete().where(analysis_result.c.analysis_result_id == analysis_result_id))
        analysis_result_parse_service.remove_analysis_result_by_analsyis_result_id(analysis_id,analysis_result_id)
    return {"message":"success"}



@sample_result.post(
    "/fast-api/add-sample-analysis",
    tags=['analsyis_result'],
    description="根据项目名称导入样本")
async def add_sample_analysis(project):
    insert_sample_list = []
    with get_engine().begin() as conn:
        stmt = samples.select().where(samples.c.project==project)
        result  = conn.execute(stmt).fetchall()
        sample_list = [dict(row._mapping) for row in result]
        # sample_list = sample_list[1:2]
        
        for item in sample_list:
            insert_sample_list.append({
                "sample_key":item['sample_key'],
                "sample_name":item['sample_name'],
                "analysis_method":f"V1_{item['sample_composition']}_{item['sequencing_technique']}_{item['sequencing_target']}",
                "content":json.dumps({
                    "fastq1":item['fastq1'],
                    "fastq2":item['fastq2']
                }),
                "project":item['project'],
                "content_type":"json",
            })
    with get_db_session() as db:
        for item in insert_sample_list:
            analysisResult = db.query(SampleAnalysisResult) \
                    .filter(and_(SampleAnalysisResult.sample_key == item['sample_key'], \
                        SampleAnalysisResult.analysis_method == item['analysis_method'] \
                        )).first()
            if analysisResult:
                analysisResult.sample_name = item['sample_name']
                analysisResult.analysis_method = item['analysis_method']
                analysisResult.content = item['content']
                analysisResult.project = item['project']
                analysisResult.content_type = item['content_type']
            else:
                analysisResult = SampleAnalysisResult(**item)
                db.add(analysisResult)
            db.commit()
            db.refresh(analysisResult)

        return {"message":"success"}

    # component_id:str
    # project: str
    # analysis_method: str
    # content: str
    # analysis_key: str


@sample_result.post("/import-data",tags=['analsyis_result'])
async def import_data(importDataList:List[ImportData]):
    with get_engine().begin() as conn:
        for importData in importDataList:
            if not importData.file_name:
                importData.file_name = importData.sample_name
                
            find_sample = sample_service.find_by_sample_name_and_project(conn,importData.sample_name,importData.project)
            sample_id = None
            if  find_sample:
                sample_id = find_sample.sample_id    
            else:
                sample_id = str(uuid.uuid4())
                sample_service.add_sample(conn,{"sample_name":importData.sample_name,"sample_id":sample_id,"project":importData.project}) 
            

            stmt = analysis_result.select().where(and_(
                analysis_result.c.sample_id==sample_id,
                analysis_result.c.component_id==importData.component_id,
                analysis_result.c.project==importData.project,
                analysis_result.c.sample_source==importData.sample_source
            ))
            result = conn.execute(stmt).fetchall()
            if result:
                raise HTTPException(status_code=500, detail=f"分析结果{importData.sample_name}已存在!")

            analysis_result_id = str(uuid.uuid4())
            stmt = analysis_result.insert().values(
                component_id=importData.component_id,
                project=importData.project,
                # analysis_method=analysis_method,
                content=importData.content,
                sample_id=sample_id,
                analysis_result_id=analysis_result_id,
                file_name=importData.file_name,
                sample_source=importData.sample_source,
                # sample_name=importData.sample_name,
                # sample_name=analysis_label,
                content_type="json",
                analysis_type="import_data"
            )
            conn.execute(stmt)
    return {"message":"success"}





@sample_result.post("/parse-import-data",tags=['analsyis_result'])
async def parse_import_data(parseImportData:ParseImportData):

    content = parseImportData.content
    content = json.loads(content)
    result = from_glob_get_file(content)
    # for item in result:
    #     item['file_name'] = item['analysis_key']
    return result

@sample_result.post("/analysis-result/update-analsyis-result",tags=['analsyis_result'])
def update_analsyis_result(updateAnalysisResult:UpdateAnalysisResult):
    with get_engine().begin() as conn:
        stmt = analysis_result.update() 
        stmt = stmt.where(analysis_result.c.id==updateAnalysisResult.id)
        stmt = stmt.values(updateAnalysisResult.model_dump())
        conn.execute(stmt)
        conn.commit()
    return {"message":"success"}


@sample_result.post("/sample/add-sample-metadata",tags=['sample'])
async def add_sample_metadata(sample_metadata:AddSampleMetadata ):
    with get_engine().begin() as conn:
        sample_id = str(uuid.uuid4())
        data = {k:v for k,v in sample_metadata.model_dump().items() if v is not None and k!="sample_id" and k!="analysis_result_id"}

        if sample_metadata.analysis_result_id:
            analysis_result = analysis_result_service.find_by_analysis_result_id(conn,sample_metadata.analysis_result_id)
            if not analysis_result:
                raise HTTPException(status_code=500, detail=f"分析结果{sample_metadata.analysis_result_id}不存在!")
    
            stmt = samples.select().where(samples.c.sample_id==analysis_result.sample_id)
            result = conn.execute(stmt).mappings().first()
            if result:
                raise HTTPException(status_code=500, detail=f"样本metadata{result.sample_id}已存在!")
            analysis_result_service.update_sample_id(conn,sample_metadata.analysis_result_id,sample_id)
            data["project"] = analysis_result.project

        if not sample_metadata.project:
            raise HTTPException(status_code=500, detail=f"项目不能为空!")
        
        data["sample_id"] = sample_id
        stmt = samples.insert().values(data)
        conn.execute(stmt)

    return {"message":"success"}

@sample_result.post("/sample/update-sample-metadata",tags=['sample'])
async def update_sample_metadata(sample_metadata:UpdateSampleMetadata    ):
    data = sample_metadata.model_dump()
    data = {k:v for k,v in data.items() if v is not None and k!="sample_id" }
    with get_engine().begin() as conn:
        stmt = samples.update().where(samples.c.sample_id==sample_metadata.sample_id).values(data)
        conn.execute(stmt)
        conn.commit()
    return {"message":"success"}

@sample_result.post("/sample/update-sample-metadata-list",tags=['sample'])
async def update_sample_metadata(sample_metadata_list:list[UpdateSampleMetadata]):
    for sample_metadata in sample_metadata_list:
        data = sample_metadata.model_dump()
        data = {k:v for k,v in data.items() if v is not None and k!="sample_id" }
        with get_engine().begin() as conn:
            stmt = samples.update().where(samples.c.sample_id==sample_metadata.sample_id).values(data)
            conn.execute(stmt)
            # conn.commit()
    return {"message":"success"}
@sample_result.get("/sample/find-sample-metadata-by-id/{sample_id}",tags=['sample'])
async def find_sample_metadata_by_id(sample_id:str):
    with get_engine().begin() as conn:
        stmt = samples.select().where(samples.c.sample_id==sample_id)
        result = conn.execute(stmt).mappings().first()
    return result


@sample_result.delete("/sample/delete-sample-by-sample-id/{sample_id}",tags=['sample'])
async def delete_sample_by_sample_id(sample_id:str):
    with get_engine().begin() as conn:
        stmt = samples.delete().where(samples.c.sample_id==sample_id)
        conn.execute(stmt)
        conn.commit()
    return {"message":"success"}


@sample_result.post("/sample/bind-sample-to-analysis-result",tags=['sample'])
async def bind_sample_to_analysis_result(bindSample:BindSample):
    with get_engine().begin() as conn:
        stmt = samples.select().where(samples.c.sample_id==bindSample.sample_id)
        result = conn.execute(stmt).mappings().first()
        if not result:
            raise HTTPException(status_code=500, detail=f"样本metadata{bindSample.sample_id}不存在!")
        analysis_result = analysis_result_service.find_by_analysis_result_id(conn,bindSample.analysis_result_id)
        if not analysis_result:
            raise HTTPException(status_code=500, detail=f"分析结果{bindSample.analysis_result_id}不存在!")
        analysis_result_service.update_sample_id(conn,bindSample.analysis_result_id,bindSample.sample_id)
    return {"message":"success"}


import os
from watchfiles import awatch
from importlib.resources import files
from importlib import import_module
from brave.api.config.db import get_engine
from brave.api.core.event import WatchFileEvent
from brave.api.models.core import analysis
import asyncio
import inspect
import logging
from brave.api.service.sse_service import   SSESessionService
from brave.api.service.analysis_result_parse import AnalysisResultParse
from brave.api.service.listener_files_service import ListenerFilesService
from brave.api.core.routers.watch_file_event_router import WatchFileEvenetRouter
from brave.app_container import AppContainer
from dependency_injector.wiring import Provide

# logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class FileWatcherService:
    def __init__(
        self, 
        watch_path: str,
        watchfile_event_router:WatchFileEvenetRouter=Provide[AppContainer.watchfile_event_router]
        # sse_service: SSESessionService, 
        # analysis_result_parse_service: AnalysisResultParse,
        # listener_files_service: ListenerFilesService
    ):
        """
        初始化 FileWatcher 实例。
        
        :param watch_path: 要监控的文件夹路径
        :param listener_prefix: 监听器文件名前缀，默认是 "file"
        """
        self.watch_path = watch_path
        self.watchfile_event_router = watchfile_event_router
        # self.sse_service = sse_service
        # self.analysis_result_parse_service = analysis_result_parse_service
        # self.listener_files_service = listener_files_service

    # def _load_listener_files(self):
    #     """加载所有符合条件的文件监听器"""
    #     listener_files = files("brave.api.listener")
    #     return [
    #         item.stem
    #         for item in listener_files.iterdir()
    #         if item.is_file() and item.name.endswith(".py")
    #         and item.name != "__init__.py" and item.name.startswith(self.listener_prefix)
    #     ]

    # async def execute_listener(self, func, args):
    #     """执行文件变更后的监听器函数"""
    #     if isinstance(self.listener_files, list) and len(self.listener_files) > 0:
    #         for name in self.listener_files:
    #             full_module = f"brave.api.listener.{name}"
    #             mod = import_module(full_module)
    #             if hasattr(mod, func):
    #                 run_func = getattr(mod, func)
    #                 if inspect.iscoroutinefunction(run_func):
    #                     # 如果是异步函数，创建一个异步任务
    #                     await run_func(**args)
    #                 else:
    #                     # 否则通过线程池执行
    #                     await asyncio.to_thread(run_func, **args)

    async def watch_folder(self):
        """文件变更监控任务"""
        print(f"开始监控文件夹: {self.watch_path}")
        async for changes in awatch(self.watch_path, recursive=False, step=3000):

            for change, file_path in changes:
                event:WatchFileEvent
                if "trace" in file_path:
                    event = WatchFileEvent.TRACE_LOG
                    analysis_id = os.path.basename(file_path).replace(".trace.log","")
                elif "workflow" in file_path:
                    event =WatchFileEvent.WORKFLOW_LOG
                    analysis_id = os.path.basename(file_path).replace(".workflow.log","")   
                else:
                    continue
                await self.watchfile_event_router.dispatch(event,{"event_type":event.value,"file_path":file_path,"analysis_id":analysis_id})   
                # # 触发文件变更事件
                # await self.listener_files_service.execute_listener("file_change",
                #  {"change": change, 
                #  "file_path": file_path, 
                #  "sse_service": 
                #  self.sse_service,
                #  "analysis_result_parse_service":self.analysis_result_parse_service})


# 示例：如何使用 FileWatcher 类
# async def main():
#     # 创建 FileWatcher 实例并开始监控
#     file_watcher = FileWatcher(watch_path="path_to_your_folder")
#     await file_watcher.watch_folder()

# 在 FastAPI 或其他异步环境中调用
# asyncio.run(main())  # 如果直接运行，可以这样调用



from collections import defaultdict
from typing import Callable, Awaitable, Dict, List, Any, Union, Coroutine
import asyncio

Callback = Union[Callable[[dict], None], Callable[[dict], Coroutine]]

class PubSubManager:
    def __init__(self):
        self.subscribers: dict[str, set[Callback]] = defaultdict(set)

    def subscribe(self, topic: str, callback: Callback):
        self.subscribers[topic].add(callback)

    def unsubscribe(self, topic: str, callback: Callback):
        self.subscribers[topic].discard(callback)

    async def publish(self, topic: str, message: dict):
        for cb in list(self.subscribers.get(topic, [])):
            try:
                if asyncio.iscoroutinefunction(cb):
                    await cb(message)
                else:
                    cb(message)
            except Exception as e:
                print(f"[PubSub ERROR] {e}")




from collections import defaultdict

from pydantic import BaseModel
from brave.api.config.db import get_engine
from brave.api.core.evenet_bus import EventBus
import asyncio


from .result_parse import ResultParse
from concurrent.futures import ThreadPoolExecutor, as_completed


class AnalysisManage:
    def __init__(
        self,
        event_bus:EventBus) -> None:
        self._result_parse: dict[str, ResultParse] = defaultdict()
        self.event_bus = event_bus
        self.pool = ThreadPoolExecutor(max_workers=3)
    
    def create(self,analysis_id:str):

        self._result_parse[analysis_id] = ResultParse(analysis_id,self.event_bus)
    
    async def parse(self,analysis_id):
        # pool.submit(self._result_parse[analysis_id].parse(), f"任务-{i}")
        # await self._result_parse[analysis_id].parse()
        loop = asyncio.get_running_loop()
        _result_parse = ResultParse(analysis_id,self.event_bus)
        loop.run_in_executor(
            self.pool,
            lambda: asyncio.run(_result_parse.parse())
        )

    def remove(self,analysis_id:str):
        del self._result_parse[analysis_id]